# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_service import fixture as service_fixture

from nova import test
from nova.virt.disk.mount import api
from nova.virt.disk.mount import block
from nova.virt.disk.mount import loop
from nova.virt.disk.mount import nbd
from nova.virt.image import model as imgmodel


PARTITION = 77
ORIG_DEVICE = "/dev/null"
AUTOMAP_PARTITION = "/dev/nullp77"
MAP_PARTITION = "/dev/mapper/nullp77"


class MountTestCase(test.NoDBTestCase):
    def setUp(self):
        super(MountTestCase, self).setUp()
        # Make RetryDecorator not actually sleep on retries
        self.useFixture(service_fixture.SleepFixture())

    def _test_map_dev(self, partition):
        mount = api.Mount(mock.sentinel.image, mock.sentinel.mount_dir)
        mount.device = ORIG_DEVICE
        mount.partition = partition
        mount.map_dev()
        return mount

    def _exists_effect(self, data):
        def exists_effect(filename):
            try:
                v = data[filename]
                if isinstance(v, list):
                    if len(v) > 0:
                        return v.pop(0)
                    self.fail("Out of items for: %s" % filename)
                return v
            except KeyError:
                self.fail("Unexpected call with: %s" % filename)
        return exists_effect

    def _check_calls(self, exists, filenames, trailing=0):
        self.assertEqual([mock.call(x) for x in filenames],
                exists.call_args_list[:len(filenames)])
        self.assertEqual([mock.call(MAP_PARTITION)] * trailing,
            exists.call_args_list[len(filenames):])

    @mock.patch('os.path.exists')
    def test_map_dev_partition_search(self, exists):
        exists.side_effect = self._exists_effect({
            ORIG_DEVICE: True})
        mount = self._test_map_dev(-1)
        self._check_calls(exists, [ORIG_DEVICE])
        self.assertNotEqual("", mount.error)
        self.assertFalse(mount.mapped)

    @mock.patch('os.path.exists')
    @mock.patch('nova.privsep.fs.create_device_maps',
                return_value=(None, None))
    def test_map_dev_good(self, mock_create_maps, mock_exists):
        mock_exists.side_effect = self._exists_effect({
            ORIG_DEVICE: True,
            AUTOMAP_PARTITION: False,
            MAP_PARTITION: [False, True]})
        mount = self._test_map_dev(PARTITION)
        self._check_calls(mock_exists, [ORIG_DEVICE, AUTOMAP_PARTITION], 2)
        self.assertEqual("", mount.error)
        self.assertTrue(mount.mapped)

    @mock.patch('os.path.exists')
    @mock.patch('nova.privsep.fs.create_device_maps',
                return_value=(None, None))
    def test_map_dev_error(self, mock_create_maps, mock_exists):
        mock_exists.side_effect = self._exists_effect({
            ORIG_DEVICE: True,
            AUTOMAP_PARTITION: False,
            MAP_PARTITION: False})
        mount = self._test_map_dev(PARTITION)
        self._check_calls(mock_exists, [ORIG_DEVICE, AUTOMAP_PARTITION],
                api.MAX_FILE_CHECKS + 1)
        self.assertNotEqual("", mount.error)
        self.assertFalse(mount.mapped)

    @mock.patch('os.path.exists')
    @mock.patch('nova.privsep.fs.create_device_maps',
                return_value=(None, None))
    def test_map_dev_error_then_pass(self, mock_create_maps, mock_exists):
        mock_exists.side_effect = self._exists_effect({
            ORIG_DEVICE: True,
            AUTOMAP_PARTITION: False,
            MAP_PARTITION: [False, False, True]})
        mount = self._test_map_dev(PARTITION)
        self._check_calls(mock_exists, [ORIG_DEVICE, AUTOMAP_PARTITION], 3)
        self.assertEqual("", mount.error)
        self.assertTrue(mount.mapped)

    @mock.patch('os.path.exists')
    def test_map_dev_automap(self, exists):
        exists.side_effect = self._exists_effect({
            ORIG_DEVICE: True,
            AUTOMAP_PARTITION: True})
        mount = self._test_map_dev(PARTITION)
        self._check_calls(exists,
            [ORIG_DEVICE, AUTOMAP_PARTITION, AUTOMAP_PARTITION])
        self.assertEqual(AUTOMAP_PARTITION, mount.mapped_device)
        self.assertTrue(mount.automapped)
        self.assertTrue(mount.mapped)

    @mock.patch('os.path.exists')
    def test_map_dev_else(self, exists):
        exists.side_effect = self._exists_effect({
            ORIG_DEVICE: True,
            AUTOMAP_PARTITION: True})
        mount = self._test_map_dev(None)
        self._check_calls(exists, [ORIG_DEVICE])
        self.assertEqual(ORIG_DEVICE, mount.mapped_device)
        self.assertFalse(mount.automapped)
        self.assertTrue(mount.mapped)

    def test_instance_for_format_raw(self):
        image = imgmodel.LocalFileImage("/some/file.raw",
                                        imgmodel.FORMAT_RAW)
        mount_dir = '/mount/dir'
        partition = -1
        inst = api.Mount.instance_for_format(image, mount_dir, partition)
        self.assertIsInstance(inst, loop.LoopMount)

    def test_instance_for_format_qcow2(self):
        image = imgmodel.LocalFileImage("/some/file.qcows",
                                        imgmodel.FORMAT_QCOW2)
        mount_dir = '/mount/dir'
        partition = -1
        inst = api.Mount.instance_for_format(image, mount_dir, partition)
        self.assertIsInstance(inst, nbd.NbdMount)

    def test_instance_for_format_block(self):
        image = imgmodel.LocalBlockImage(
            "/dev/mapper/instances--instance-0000001_disk",)
        mount_dir = '/mount/dir'
        partition = -1
        inst = api.Mount.instance_for_format(image, mount_dir, partition)
        self.assertIsInstance(inst, block.BlockMount)

    def test_instance_for_device_loop(self):
        image = mock.MagicMock()
        mount_dir = '/mount/dir'
        partition = -1
        device = '/dev/loop0'
        inst = api.Mount.instance_for_device(image, mount_dir, partition,
                                               device)
        self.assertIsInstance(inst, loop.LoopMount)

    def test_instance_for_device_loop_partition(self):
        image = mock.MagicMock()
        mount_dir = '/mount/dir'
        partition = 1
        device = '/dev/mapper/loop0p1'
        inst = api.Mount.instance_for_device(image, mount_dir, partition,
                                               device)
        self.assertIsInstance(inst, loop.LoopMount)

    def test_instance_for_device_nbd(self):
        image = mock.MagicMock()
        mount_dir = '/mount/dir'
        partition = -1
        device = '/dev/nbd0'
        inst = api.Mount.instance_for_device(image, mount_dir, partition,
                                               device)
        self.assertIsInstance(inst, nbd.NbdMount)

    def test_instance_for_device_nbd_partition(self):
        image = mock.MagicMock()
        mount_dir = '/mount/dir'
        partition = 1
        device = '/dev/mapper/nbd0p1'
        inst = api.Mount.instance_for_device(image, mount_dir, partition,
                                               device)
        self.assertIsInstance(inst, nbd.NbdMount)

    def test_instance_for_device_block(self):
        image = mock.MagicMock()
        mount_dir = '/mount/dir'
        partition = -1
        device = '/dev/mapper/instances--instance-0000001_disk'
        inst = api.Mount.instance_for_device(image, mount_dir, partition,
                                               device)
        self.assertIsInstance(inst, block.BlockMount)

    def test_instance_for_device_block_partiton(self):
        image = mock.MagicMock()
        mount_dir = '/mount/dir'
        partition = 1
        device = '/dev/mapper/instances--instance-0000001_diskp1'
        inst = api.Mount.instance_for_device(image, mount_dir, partition,
                                               device)
        self.assertIsInstance(inst, block.BlockMount)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from nova import context as nova_context
from nova import exception
from nova import objects
from nova import test
from nova.tests import fixtures as nova_fixtures
from nova.tests.functional import integrated_helpers


class TestDeleteWhileBooting(test.TestCase,
                             integrated_helpers.InstanceHelperMixin):
    """This tests race scenarios where an instance is deleted while booting.

    In these scenarios, the nova-api service is racing with nova-conductor
    service; nova-conductor is in the middle of booting the instance when
    nova-api begins fulfillment of a delete request. As the two services
    delete records out from under each other, both services need to handle
    it properly such that a delete request will always be fulfilled.

    Another scenario where two requests can race and delete things out from
    under each other is if two or more delete requests are racing while the
    instance is booting.

    In order to force things into states where bugs have occurred, we must
    mock some object retrievals from the database to simulate the different
    points at which a delete request races with a create request or another
    delete request. We aim to mock only the bare minimum necessary to recreate
    the bug scenarios.
    """

    def setUp(self):
        super(TestDeleteWhileBooting, self).setUp()
        self.useFixture(nova_fixtures.RealPolicyFixture())
        self.useFixture(nova_fixtures.NeutronFixture(self))
        self.useFixture(nova_fixtures.GlanceFixture(self))

        api_fixture = self.useFixture(nova_fixtures.OSAPIFixture(
            api_version='v2.1'))
        self.api = api_fixture.api

        self.ctxt = nova_context.get_context()

        # We intentionally do not start a conductor or scheduler service, since
        # our goal is to simulate an instance that has not been scheduled yet.

        # Kick off a server create request and move on once it's in the BUILD
        # state. Since we have no conductor or scheduler service running, the
        # server will "hang" in an unscheduled state for testing.
        self.server = self._create_server(expected_state='BUILD')
        # Simulate that a different request has deleted the build request
        # record after this delete request has begun processing. (The first
        # lookup of the build request occurs in the servers API to get the
        # instance object in order to delete it).
        # We need to get the build request now before we mock the method.
        self.br = objects.BuildRequest.get_by_instance_uuid(
            self.ctxt, self.server['id'])

    @mock.patch('nova.objects.build_request.BuildRequest.get_by_instance_uuid')
    def test_build_request_and_instance_not_found(self, mock_get_br):
        """This tests a scenario where another request has deleted the build
        request record and the instance record ahead of us.
        """
        # The first lookup at the beginning of the delete request in the
        # ServersController succeeds and the second lookup to handle "delete
        # while booting" in compute/api fails after a different request has
        # deleted it.
        br_not_found = exception.BuildRequestNotFound(uuid=self.server['id'])
        mock_get_br.side_effect = [self.br, br_not_found, br_not_found]
        self._delete_server(self.server)

    @mock.patch('nova.objects.build_request.BuildRequest.get_by_instance_uuid')
    @mock.patch('nova.objects.InstanceMapping.get_by_instance_uuid')
    @mock.patch('nova.objects.instance.Instance.get_by_uuid')
    def test_deleting_instance_at_the_same_time(self, mock_get_i, mock_get_im,
                                                mock_get_br):
        """This tests the scenario where another request is trying to delete
        the instance record at the same time we are, while the instance is
        booting. An example of this: while the create and delete are running at
        the same time, the delete request deletes the build request, the create
        request finds the build request already deleted when it tries to delete
        it. The create request deletes the instance record and then delete
        request tries to lookup the instance after it deletes the build
        request. Its attempt to lookup the instance fails because the create
        request already deleted it.
        """
        # First lookup at the beginning of the delete request in the
        # ServersController succeeds, second lookup to handle "delete while
        # booting" in compute/api fails after the conductor has deleted it.
        br_not_found = exception.BuildRequestNotFound(uuid=self.server['id'])
        mock_get_br.side_effect = [self.br, br_not_found]
        # Simulate the instance transitioning from having no cell assigned to
        # having a cell assigned while the delete request is being processed.
        # First lookup of the instance mapping has the instance unmapped (no
        # cell) and subsequent lookups have the instance mapped to cell1.
        no_cell_im = objects.InstanceMapping(
            context=self.ctxt, instance_uuid=self.server['id'],
            cell_mapping=None)
        has_cell_im = objects.InstanceMapping(
            context=self.ctxt, instance_uuid=self.server['id'],
            cell_mapping=self.cell_mappings['cell1'])
        mock_get_im.side_effect = [
            no_cell_im, has_cell_im, has_cell_im, has_cell_im, has_cell_im]
        # Simulate that the instance object has been created by the conductor
        # in the create path while the delete request is being processed.
        # First lookups are before the instance has been deleted and the last
        # lookup is after the conductor has deleted the instance. Use the build
        # request to make an instance object for testing.
        i = self.br.get_new_instance(self.ctxt)
        i_not_found = exception.InstanceNotFound(instance_id=self.server['id'])
        mock_get_i.side_effect = [i, i, i, i_not_found, i_not_found]

        # Simulate that the conductor is running instance_destroy at the same
        # time as we are.
        def fake_instance_destroy(*args, **kwargs):
            # NOTE(melwitt): This is a misleading exception, as it is not only
            # raised when a constraint on 'host' is not met, but also when two
            # instance_destroy calls are racing. In this test, the soft delete
            # returns 0 rows affected because another request soft deleted the
            # record first.
            raise exception.ObjectActionError(
                action='destroy', reason='host changed')

        self.stub_out(
            'nova.objects.instance.Instance.destroy', fake_instance_destroy)
        self._delete_server(self.server)



# Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from nova import exception
from nova.i18n import _

# Define the minimum and maximum version of the API across all of the
# REST API. The format of the version is:
# X.Y where:
#
# - X will only be changed if a significant backwards incompatible API
# change is made which affects the API as whole. That is, something
# that is only very very rarely incremented.
#
# - Y when you make any change to the API. Note that this includes
# semantic changes which may not affect the input or output formats or
# even originate in the API code layer. We are not distinguishing
# between backwards compatible and backwards incompatible changes in
# the versioning system. It must be made clear in the documentation as
# to what is a backwards compatible change and what is a backwards
# incompatible one.

#
# You must update the API version history string below with a one or
# two line description as well as update rest_api_version_history.rst
REST_API_VERSION_HISTORY = """REST API Version History:

    * 2.1 - Initial version. Equivalent to v2.0 code
    * 2.2 - Adds (keypair) type parameter for os-keypairs plugin
            Fixes success status code for create/delete a keypair method
    * 2.3 - Exposes additional os-extended-server-attributes
            Exposes delete_on_termination for os-extended-volumes
    * 2.4 - Exposes reserved field in os-fixed-ips.
    * 2.5 - Allow server search option ip6 for non-admin
    * 2.6 - Consolidate the APIs for getting remote consoles
    * 2.7 - Check flavor type before add tenant access.
    * 2.8 - Add new protocol for VM console (mks)
    * 2.9 - Exposes lock information in server details.
    * 2.10 - Allow admins to query, create and delete keypairs owned by any
             user.
    * 2.11 - Exposes forced_down attribute for os-services
    * 2.12 - Exposes VIF net_id in os-virtual-interfaces
    * 2.13 - Add project id and user id information for os-server-groups API
    * 2.14 - Remove onSharedStorage from evacuate request body and remove
             adminPass from the response body
    * 2.15 - Add soft-affinity and soft-anti-affinity policies
    * 2.16 - Exposes host_status for servers/detail and servers/{server_id}
    * 2.17 - Add trigger_crash_dump to server actions
    * 2.18 - Makes project_id optional in v2.1
    * 2.19 - Allow user to set and get the server description
    * 2.20 - Add attach and detach volume operations for instances in shelved
             and shelved_offloaded state
    * 2.21 - Make os-instance-actions read deleted instances
    * 2.22 - Add API to force live migration to complete
    * 2.23 - Add index/show API for server migrations.
             Also add migration_type for /os-migrations and add ref link for it
             when the migration is an in progress live migration.
    * 2.24 - Add API to cancel a running live migration
    * 2.25 - Make block_migration support 'auto' and remove
             disk_over_commit for os-migrateLive.
    * 2.26 - Adds support of server tags
    * 2.27 - Adds support for new-style microversion headers while
             keeping support for the original style.
    * 2.28 - Changes compute_node.cpu_info from string to object
    * 2.29 - Add a force flag in evacuate request body and change the
             behaviour for the host flag by calling the scheduler.
    * 2.30 - Add a force flag in live-migrate request body and change the
             behaviour for the host flag by calling the scheduler.
    * 2.31 - Fix os-console-auth-tokens to work for all console types.
    * 2.32 - Add tag to networks and block_device_mapping_v2 in server boot
             request body.
    * 2.33 - Add pagination support for hypervisors.
    * 2.34 - Checks before live-migration are made in asynchronous way.
             os-Migratelive Action does not throw badRequest in case of
             pre-checks failure. Verification result is available over
             instance-actions.
    * 2.35 - Adds keypairs pagination support.
    * 2.36 - Deprecates all the API which proxy to another service and fping
             API.
    * 2.37 - Adds support for auto-allocating networking, otherwise known as
             "Get me a Network". Also enforces server.networks.uuid to be in
             UUID format.
    * 2.38 - Add a condition to return HTTPBadRequest if invalid status is
             provided for listing servers.
    * 2.39 - Deprecates image-metadata proxy API
    * 2.40 - Adds simple tenant usage pagination support.
    * 2.41 - Return uuid attribute for aggregates.
    * 2.42 - In the context of device tagging at instance boot time,
             re-introduce the tag attribute that, due to bugs, was lost
             starting with version 2.33 for block devices and starting with
             version 2.37 for network interfaces.
    * 2.43 - Deprecate os-hosts API
    * 2.44 - The servers action addFixedIp, removeFixedIp, addFloatingIp,
             removeFloatingIp and os-virtual-interfaces APIs are deprecated.
    * 2.45 - The createImage and createBackup APIs no longer return a Location
             header in the response for the snapshot image, they now return a
             json dict in the response body with an image_id key and uuid
             value.
    * 2.46 - Return ``X-OpenStack-Request-ID`` header on requests.
    * 2.47 - When displaying server details, display the flavor as a dict
             rather than a link.  If the user is prevented from retrieving
             the flavor extra-specs by policy, simply omit the field from
             the output.
    * 2.48 - Standardize VM diagnostics info.
    * 2.49 - Support tagged attachment of network interfaces and block devices.
    * 2.50 - Exposes ``server_groups`` and ``server_group_members`` keys in
             GET & PUT ``os-quota-class-sets`` APIs response.
             Also filter out Network related quotas from
             ``os-quota-class-sets`` API
    * 2.51 - Adds new event name to external-events (volume-extended). Also,
             non-admins can see instance action event details except for the
             traceback field.
    * 2.52 - Adds support for applying tags when creating a server.
    * 2.53 - Service and compute node (hypervisor) database ids are hidden.
             The os-services and os-hypervisors APIs now return a uuid in the
             id field, and takes a uuid in requests. PUT and GET requests
             and responses are also changed.
    * 2.54 - Enable reset key pair while rebuilding instance.
    * 2.55 - Added flavor.description to GET/POST/PUT flavors APIs.
    * 2.56 - Add a host parameter in migrate request body in order to
             enable users to specify a target host in cold migration.
             The target host is checked by the scheduler.
    * 2.57 - Deprecated personality files from POST /servers and the rebuild
             server action APIs. Added the ability to pass new user_data to
             the rebuild server action API. Personality / file injection
             related limits and quota resources are also removed.
    * 2.58 - Add pagination support and changes-since filter for
             os-instance-actions API.
    * 2.59 - Add pagination support and changes-since filter for os-migrations
             API. And the os-migrations API now returns both the id and the
             uuid in response.
    * 2.60 - Add support for attaching a single volume to multiple instances.
    * 2.61 - Exposes flavor extra_specs in the flavor representation. Flavor
             extra_specs will be included in Response body of GET, POST, PUT
             /flavors APIs.
    * 2.62 - Add ``host`` and ``hostId`` fields to instance action detail API
             responses.
    * 2.63 - Add support for applying trusted certificates when creating or
             rebuilding a server.
    * 2.64 - Add support for the "max_server_per_host" policy rule for
             ``anti-affinity`` server group policy, the ``policies`` and
             ``metadata`` fields are removed and the ``policy`` (required)
             and ``rules`` (optional) fields are added in response body of
             GET, POST /os-server-groups APIs and GET
             /os-server-groups/{group_id} API.
    * 2.65 - Add support for abort live migrations in ``queued`` and
             ``preparing`` status.
    * 2.66 - Add ``changes-before`` to support users to specify the
             ``updated_at`` time to filter nova resources, the resources
             include the servers API, os-instance-action API and
             os-migrations API.
    * 2.67 - Adds the optional ``volume_type`` field to the
             ``block_device_mapping_v2`` parameter when creating a server.
    * 2.68 - Remove support for forced live migration and evacuate server
             actions.
    * 2.69 - Add support for returning minimal constructs for ``GET /servers``,
             ``GET /servers/detail``, ``GET /servers/{server_id}`` and
             ``GET /os-services`` when there is a transient unavailability
             condition in the deployment like an infrastructure failure.
    * 2.70 - Exposes virtual device tags in the response of the
             ``os-volume_attachments`` and ``os-interface`` APIs.
    * 2.71 - Adds the ``server_groups`` field to ``GET /servers/{id}``,
            ``PUT /servers/{server_id}`` and
            ``POST /servers/{server_id}/action`` (rebuild) responses.
    * 2.72 - Add support for neutron ports with resource request during server
             create. Server move operations are not yet supported for servers
             with such ports.
    * 2.73 - Adds support for specifying a reason when locking the server and
             exposes this via the response from ``GET /servers/detail``,
             ``GET /servers/{server_id}``, ``PUT servers/{server_id}`` and
             ``POST /servers/{server_id}/action`` where the action is rebuild.
             It also supports ``locked`` as a filter/sort parameter for
             ``GET /servers/detail`` and ``GET /servers``.
    * 2.74 - Add support for specifying ``host`` and/or ``hypervisor_hostname``
             in request body to ``POST /servers``. Allow users to specify which
             host/node they want their servers to land on and still be
             validated by the scheduler.
    * 2.75 - Multiple API cleanup listed below:
             - 400 for unknown param for query param and for request body.
             - Making server representation always consistent among GET, PUT
               and Rebuild serevr APIs response.
             - Change the default return value of swap field from the empty
               string to 0 (integer) in flavor APIs.
             - Return ``servers`` field always in the response of GET
               hypervisors API even there are no servers on hypervisor.
    * 2.76 - Adds ``power-update`` event to ``os-server-external-events`` API.
             The changes to the power state of an instance caused by this event
             can be viewed through
             ``GET /servers/{server_id}/os-instance-actions`` and
             ``GET /servers/{server_id}/os-instance-actions/{request_id}``.
    * 2.77 - Add support for specifying ``availability_zone`` to unshelve of a
             shelved offload server.
    * 2.78 - Adds new API ``GET /servers/{server_id}/topology`` which shows
             NUMA topology of a given server.
    * 2.79 - Adds support for specifying ``delete_on_termination`` field in the
             request body to
             ``POST /servers/{server_id}/os-volume_attachments`` and exposes
             this via the response from
             ``POST /servers/{server_id}/os-volume_attachments``,
             ``GET /servers/{server_id}/os-volume_attachments`` and
             ``GET /servers/{server_id}/os-volume_attachments/{volume_id}``.
    * 2.80 - Adds support for optional query parameters ``user_id`` and
             ``project_id`` to the ``GET /os-migrations`` API and exposes
             ``user_id`` and ``project_id`` via the response from
             ``GET /os-migrations``,
             ``GET /servers/{server_id}/migrations``, and
             ``GET /servers/{server_id}/migrations/{migration_id}``.
    * 2.81 - Adds support for image cache management by aggregate by adding
             ``POST /os-aggregates/{aggregate_id}/images``.
    * 2.82 - Adds ``accelerator-request-bound`` event to
             ``os-server-external-events`` API. This event is sent by Cyborg
             to indicate completion of ARQ binding. The ARQs can be obtained
             from Cyborg with ``GET /v2/accelerator_requests?instance={uuid}``
    * 2.83 - Allow more filter parameters for ``GET /servers/detail`` and
             ``GET /servers`` for non-admin.
    * 2.84 - Adds ``details`` field to instance action events.
    * 2.85 - Add support for
             ``PUT /servers/{server_id}/os-volume_attachments/{volume_id}``
             which supports specifying the ``delete_on_termination`` field in
             the request body to change the attached volume's flag.
    * 2.86 - Add support for validation of known extra specs to the
             ``POST /flavors/{flavor_id}/os-extra_specs`` and
             ``PUT /flavors/{flavor_id}/os-extra_specs/{id}`` APIs.
    * 2.87 - Adds support for rescuing boot from volume instances when the
             compute host reports the COMPUTE_RESCUE_BFV capability trait.
    * 2.88 - Drop statistics-style fields from the ``/os-hypervisors/detail``
             and ``/os-hypervisors/{hypervisor_id}`` APIs, and remove the
             ``/os-hypervisors/statistics`` and
             ``/os-hypervisors/{hypervisor_id}/uptime`` APIs entirely.
    * 2.89 - Add ``attachment_id``, ``bdm_uuid`` and remove ``id`` from the
             responses of ``GET  /servers/{server_id}/os-volume_attachments``
             and ``GET /servers/{server_id}/os-volume_attachments/{volume_id}``
    * 2.90 - Add support for requesting a specific hostname when creating,
             updating or rebuilding an instance. The
             ``OS-EXT-SRV-ATTR:hostname`` attribute is now returned in various
             server responses regardless of policy configuration.
    * 2.91 - Add support to unshelve instance to a specific host and
             to pin/unpin AZ.
    * 2.92 - Drop generation of keypair, add keypair name validation on
             ``POST /os-keypairs`` and allow including @ and dot (.) characters
             in keypair name.
    * 2.93 - Add support for volume backed server rebuild.
    * 2.94 - Allow FQDN in server hostname.
    * 2.95 - Evacuate will now stop instance at destination.
    * 2.96 - Add support for returning pinned_availability_zone in
             ``server show`` and ``server list --long`` responses.
"""

# The minimum and maximum versions of the API supported
# The default api version request is defined to be the
# minimum version of the API supported.
# Note(cyeoh): This only applies for the v2.1 API once microversions
# support is fully merged. It does not affect the V2 API.
_MIN_API_VERSION = '2.1'
_MAX_API_VERSION = '2.96'
DEFAULT_API_VERSION = _MIN_API_VERSION

# Almost all proxy APIs which are related to network, images and baremetal
# were deprecated from 2.36.
MAX_PROXY_API_SUPPORT_VERSION = '2.35'
MIN_WITHOUT_PROXY_API_SUPPORT_VERSION = '2.36'

# Starting from microversion 2.39 also image-metadata proxy API is deprecated.
MAX_IMAGE_META_PROXY_API_VERSION = '2.38'
MIN_WITHOUT_IMAGE_META_PROXY_API_VERSION = '2.39'


# NOTE(cyeoh): min and max versions declared as functions so we can
# mock them for unittests. Do not use the constants directly anywhere
# else.
def min_api_version():
    return APIVersionRequest(_MIN_API_VERSION)


def max_api_version():
    return APIVersionRequest(_MAX_API_VERSION)


def is_supported(req, min_version=_MIN_API_VERSION,
                 max_version=_MAX_API_VERSION):
    """Check if API request version satisfies version restrictions.

    :param req: request object
    :param min_version: minimal version of API needed for correct
           request processing
    :param max_version: maximum version of API needed for correct
           request processing

    :returns: True if request satisfies minimal and maximum API version
             requirements. False in other case.
    """

    return (APIVersionRequest(max_version) >= req.api_version_request >=
            APIVersionRequest(min_version))


class APIVersionRequest(object):
    """This class represents an API Version Request with convenience
    methods for manipulation and comparison of version
    numbers that we need to do to implement microversions.
    """

    def __init__(self, version_string=None):
        """Create an API version request object.

        :param version_string: String representation of APIVersionRequest.
            Correct format is 'X.Y', where 'X' and 'Y' are int values.
            None value should be used to create Null APIVersionRequest,
            which is equal to 0.0
        """
        self.ver_major = 0
        self.ver_minor = 0

        if version_string is not None:
            match = re.match(r"^([1-9]\d*)\.([1-9]\d*|0)$",
                             version_string)
            if match:
                self.ver_major = int(match.group(1))
                self.ver_minor = int(match.group(2))
            else:
                raise exception.InvalidAPIVersionString(version=version_string)

    def __str__(self):
        """Debug/Logging representation of object."""
        return ("API Version Request Major: %s, Minor: %s"
                % (self.ver_major, self.ver_minor))

    def is_null(self):
        return self.ver_major == 0 and self.ver_minor == 0

    def _format_type_error(self, other):
        return TypeError(_("'%(other)s' should be an instance of '%(cls)s'") %
                         {"other": other, "cls": self.__class__})

    def __lt__(self, other):
        if not isinstance(other, APIVersionRequest):
            raise self._format_type_error(other)

        return ((self.ver_major, self.ver_minor) <
                (other.ver_major, other.ver_minor))

    def __eq__(self, other):
        if not isinstance(other, APIVersionRequest):
            raise self._format_type_error(other)

        return ((self.ver_major, self.ver_minor) ==
                (other.ver_major, other.ver_minor))

    def __gt__(self, other):
        if not isinstance(other, APIVersionRequest):
            raise self._format_type_error(other)

        return ((self.ver_major, self.ver_minor) >
                (other.ver_major, other.ver_minor))

    def __le__(self, other):
        return self < other or self == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __ge__(self, other):
        return self > other or self == other

    def matches(self, min_version, max_version):
        """Returns whether the version object represents a version
        greater than or equal to the minimum version and less than
        or equal to the maximum version.

        @param min_version: Minimum acceptable version.
        @param max_version: Maximum acceptable version.
        @returns: boolean

        If min_version is null then there is no minimum limit.
        If max_version is null then there is no maximum limit.
        If self is null then raise ValueError
        """

        if self.is_null():
            raise ValueError
        if max_version.is_null() and min_version.is_null():
            return True
        elif max_version.is_null():
            return min_version <= self
        elif min_version.is_null():
            return self <= max_version
        else:
            return min_version <= self <= max_version

    def get_string(self):
        """Converts object to string representation which if used to create
        an APIVersionRequest object results in the same version request.
        """
        if self.is_null():
            raise ValueError
        return "%s.%s" % (self.ver_major, self.ver_minor)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from nova.api.openstack.compute import instance_usage_audit_log as iual
from nova.policies import base as base_policy
from nova.policies import instance_usage_audit_log as iual_policies
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit.policies import base


class InstanceUsageAuditLogPolicyTest(base.BasePolicyTest):
    """Test os-instance-usage-audit-log APIs policies with all possible
    context.
    This class defines the set of context with different roles
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will call the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(InstanceUsageAuditLogPolicyTest, self).setUp()
        self.controller = iual.InstanceUsageAuditLogController()
        self.req = fakes.HTTPRequest.blank('')
        self.controller.host_api.task_log_get_all = mock.MagicMock()
        self.controller.host_api.service_get_all = mock.MagicMock()

        # With legacy rule, all admin_api will be able to get instance usage
        # audit log.
        self.admin_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context]

    def test_show_policy(self):
        rule_name = iual_policies.BASE_POLICY_NAME % 'show'
        self.common_policy_auth(self.admin_authorized_contexts,
                                rule_name, self.controller.show,
                                self.req, '2020-03-25 14:40:00')

    def test_index_policy(self):
        rule_name = iual_policies.BASE_POLICY_NAME % 'list'
        self.common_policy_auth(self.admin_authorized_contexts,
                                rule_name, self.controller.index,
                                self.req)


class InstanceUsageNoLegacyNoScopeTest(InstanceUsageAuditLogPolicyTest):
    """Test Instance Usage API policies with deprecated rules
    disabled, but scope checking still disabled.
    """

    without_deprecated_rules = True
    rules_without_deprecation = {
        iual_policies.BASE_POLICY_NAME % 'list':
            base_policy.ADMIN,
        iual_policies.BASE_POLICY_NAME % 'show':
            base_policy.ADMIN,
    }


class InstanceUsageScopeTypePolicyTest(InstanceUsageAuditLogPolicyTest):
    """Test os-instance-usage-audit-log APIs policies with system scope
    enabled.
    This class set the nova.conf [oslo_policy] enforce_scope to True
    so that we can switch on the scope checking on oslo policy side.
    It defines the set of context with scoped token
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will run the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(InstanceUsageScopeTypePolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")

        # Scope checks remove project users power.
        self.admin_authorized_contexts = [
            self.legacy_admin_context,
            self.project_admin_context]


class InstanceUsageScopeTypeNoLegacyPolicyTest(
        InstanceUsageScopeTypePolicyTest):
    """Test Instance Usage Audit Log APIs policies with system scope enabled,
    and no more deprecated rules.
    """
    without_deprecated_rules = True
    rules_without_deprecation = {
        iual_policies.BASE_POLICY_NAME % 'list':
            base_policy.ADMIN,
        iual_policies.BASE_POLICY_NAME % 'show':
            base_policy.ADMIN,
    }



# Copyright (c) 2013 Akira Yoshiyama <akirayoshiyama at gmail dot com>
#
# This is derived from test_db_servicegroup.py.
# Copyright 2012 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import iso8601
from unittest import mock

from nova import servicegroup
from nova import test
from oslo_utils import timeutils


class MemcachedServiceGroupTestCase(test.NoDBTestCase):

    @mock.patch('nova.cache_utils.get_memcached_client')
    def setUp(self, mgc_mock):
        super(MemcachedServiceGroupTestCase, self).setUp()
        self.mc_client = mock.MagicMock()
        mgc_mock.return_value = self.mc_client
        self.flags(servicegroup_driver='mc')
        self.servicegroup_api = servicegroup.API()

    def test_is_up(self):
        service_ref = {
            'host': 'fake-host',
            'topic': 'compute'
        }
        self.mc_client.get.return_value = None

        self.assertFalse(self.servicegroup_api.service_is_up(service_ref))
        self.mc_client.get.assert_called_once_with('compute:fake-host')
        self.mc_client.reset_mock()
        self.mc_client.get.return_value = True
        self.assertTrue(self.servicegroup_api.service_is_up(service_ref))
        self.mc_client.get.assert_called_once_with('compute:fake-host')

    def test_join(self):
        service = mock.MagicMock(report_interval=1)

        self.servicegroup_api.join('fake-host', 'fake-topic', service)
        fn = self.servicegroup_api._driver._report_state
        service.tg.add_timer_args.assert_called_once_with(
            1, fn, args=[service], initial_delay=5)

    def test_report_state(self):
        service_ref = {
            'host': 'fake-host',
            'topic': 'compute'
        }
        service = mock.MagicMock(model_disconnected=False,
                                 service_ref=service_ref)
        fn = self.servicegroup_api._driver._report_state
        fn(service)
        self.mc_client.set.assert_called_once_with('compute:fake-host',
                                                   mock.ANY)

    def test_get_updated_time(self):
        updated_at_time = timeutils.parse_strtime("2016-04-18T02:56:25.198871")
        service_ref = {
            'host': 'fake-host',
            'topic': 'compute',
            'updated_at': updated_at_time.replace(tzinfo=iso8601.UTC)
        }

        # If no record returned from the mc, return record from DB
        self.mc_client.get.return_value = None
        self.assertEqual(service_ref['updated_at'],
                         self.servicegroup_api.get_updated_time(service_ref))
        self.mc_client.get.assert_called_once_with('compute:fake-host')
        # If the record in mc is newer than DB, return record from mc
        self.mc_client.reset_mock()
        retval = timeutils.utcnow()
        self.mc_client.get.return_value = retval
        self.assertEqual(retval.replace(tzinfo=iso8601.UTC),
                         self.servicegroup_api.get_updated_time(service_ref))
        self.mc_client.get.assert_called_once_with('compute:fake-host')
        # If the record in DB is newer than mc, return record from DB
        self.mc_client.reset_mock()
        service_ref['updated_at'] = \
            retval.replace(tzinfo=iso8601.UTC)
        self.mc_client.get.return_value = updated_at_time
        self.assertEqual(service_ref['updated_at'],
                         self.servicegroup_api.get_updated_time(service_ref))
        self.mc_client.get.assert_called_once_with('compute:fake-host')
        # If no record returned from the DB, return the record from mc
        self.mc_client.reset_mock()
        service_ref['updated_at'] = None
        self.mc_client.get.return_value = updated_at_time
        self.assertEqual(updated_at_time.replace(tzinfo=iso8601.UTC),
                         self.servicegroup_api.get_updated_time(service_ref))
        self.mc_client.get.assert_called_once_with('compute:fake-host')



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

guestfs_group = cfg.OptGroup('guestfs',
                             title='Guestfs Options',
                             help="""
libguestfs is a set of tools for accessing and modifying virtual
machine (VM) disk images. You can use this for viewing and editing
files inside guests, scripting changes to VMs, monitoring disk
used/free statistics, creating guests, P2V, V2V, performing backups,
cloning VMs, building VMs, formatting disks and resizing disks.
""")

enable_guestfs_debug_opts = [
    cfg.BoolOpt('debug',
                default=False,
                help="""
Enable/disables guestfs logging.

This configures guestfs to debug messages and push them to OpenStack
logging system. When set to True, it traces libguestfs API calls and
enable verbose debug messages. In order to use the above feature,
"libguestfs" package must be installed.

Related options:

Since libguestfs access and modifies VM's managed by libvirt, below options
should be set to give access to those VM's.

* ``libvirt.inject_key``
* ``libvirt.inject_partition``
* ``libvirt.inject_password``
""")
]


def register_opts(conf):
    conf.register_group(guestfs_group)
    conf.register_opts(enable_guestfs_debug_opts,
                       group=guestfs_group)


def list_opts():
    return {guestfs_group: enable_guestfs_debug_opts}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from nova.conductor import api as conductor_api
from nova import context as nova_context
from nova import objects
from nova import test
from nova.tests import fixtures as nova_fixtures
from nova.tests.functional import fixtures as func_fixtures
from nova.tests.functional import integrated_helpers


class MissingReqSpecInstanceGroupUUIDTestCase(
        test.TestCase, integrated_helpers.InstanceHelperMixin):
    """Regression recreate test for bug 1830747

    Before change I4244f7dd8fe74565180f73684678027067b4506e in Stein, when
    a cold migration would reschedule to conductor it would not send the
    RequestSpec, only the filter_properties. The filter_properties contain
    a primitive version of the instance group information from the RequestSpec
    for things like the group members, hosts and policies, but not the uuid.
    When conductor is trying to reschedule the cold migration without a
    RequestSpec, it builds a RequestSpec from the components it has, like the
    filter_properties. This results in a RequestSpec with an instance_group
    field set but with no uuid field in the RequestSpec.instance_group.
    That RequestSpec gets persisted and then because of change
    Ie70c77db753711e1449e99534d3b83669871943f, later attempts to load the
    RequestSpec from the database will fail because of the missing
    RequestSpec.instance_group.uuid.

    This test recreates the regression scenario by cold migrating a server
    to a host which fails and triggers a reschedule but without the RequestSpec
    so a RequestSpec is created/updated for the instance without the
    instance_group.uuid set which will lead to a failure loading the
    RequestSpec from the DB later.
    """

    def setUp(self):
        super(MissingReqSpecInstanceGroupUUIDTestCase, self).setUp()
        # Stub out external dependencies.
        self.useFixture(nova_fixtures.NeutronFixture(self))
        self.useFixture(nova_fixtures.GlanceFixture(self))
        self.useFixture(func_fixtures.PlacementFixture())
        # Configure the API to allow resizing to the same host so we can keep
        # the number of computes down to two in the test.
        self.flags(allow_resize_to_same_host=True)
        # Start nova controller services.
        api_fixture = self.useFixture(nova_fixtures.OSAPIFixture(
            api_version='v2.1'))
        self.api = api_fixture.admin_api
        self.start_service('conductor')
        # Use a custom weigher to make sure that we have a predictable
        # scheduling sort order.
        self.useFixture(nova_fixtures.HostNameWeigherFixture())
        self.start_service('scheduler')
        # Start two computes, one where the server will be created and another
        # where we'll cold migrate it.
        self._start_compute('host1')
        self._start_compute('host2')

    def test_cold_migrate_reschedule(self):
        # Create an anti-affinity group for the server.
        body = {
            'server_group': {
                'name': 'test-group',
                'policies': ['anti-affinity']
            }
        }
        group_id = self.api.api_post(
            '/os-server-groups', body).body['server_group']['id']

        # Create a server in the group which should land on host1 due to our
        # custom weigher.
        body = {'server': self._build_server()}
        body['os:scheduler_hints'] = {'group': group_id}
        server = self.api.post_server(body)
        server = self._wait_for_state_change(server, 'ACTIVE')
        self.assertEqual('host1', server['OS-EXT-SRV-ATTR:host'])

        # Verify the group uuid is set in the request spec.
        ctxt = nova_context.get_admin_context()
        reqspec = objects.RequestSpec.get_by_instance_uuid(ctxt, server['id'])
        self.assertEqual(group_id, reqspec.instance_group.uuid)

        # Stub out the ComputeTaskAPI.resize_instance method to not pass the
        # request spec from compute to conductor during a reschedule.
        # This simulates the pre-Stein reschedule behavior.
        original_resize_instance = conductor_api.ComputeTaskAPI.resize_instance

        def stub_resize_instance(_self, context, instance, scheduler_hint,
                                 flavor, *args, **kwargs):
            # Only remove the request spec if we know we're rescheduling
            # which we can determine from the filter_properties retry dict.
            filter_properties = scheduler_hint['filter_properties']
            if filter_properties.get('retry', {}).get('exc'):
                # Assert the group_uuid is passed through the filter properties
                self.assertIn('group_uuid', filter_properties)
                self.assertEqual(group_id, filter_properties['group_uuid'])
                kwargs.pop('request_spec', None)
            return original_resize_instance(
                _self, context, instance, scheduler_hint, flavor, *args,
                **kwargs)
        self.stub_out('nova.conductor.api.ComputeTaskAPI.resize_instance',
                      stub_resize_instance)

        # Now cold migrate the server. Because of allow_resize_to_same_host and
        # the weigher, the scheduler will pick host1 first. The FakeDriver
        # actually allows migrating to the same host so we need to stub that
        # out so the compute will raise UnableToMigrateToSelf like when using
        # the libvirt driver.
        host1_driver = self.computes['host1'].driver
        with mock.patch.dict(host1_driver.capabilities,
                             supports_migrate_to_same_host=False):
            self.api.post_server_action(server['id'], {'migrate': None})
            server = self._wait_for_state_change(server, 'VERIFY_RESIZE')
            self.assertEqual('host2', server['OS-EXT-SRV-ATTR:host'])

        # The RequestSpec.instance_group.uuid should still be set.
        reqspec = objects.RequestSpec.get_by_instance_uuid(ctxt, server['id'])
        self.assertEqual(group_id, reqspec.instance_group.uuid)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add encryption fields to BlockDeviceMapping

Revision ID: ccb0fa1a2252
Revises: 16f1fbcab42b
Create Date: 2022-01-12 15:22:47.524285
"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'ccb0fa1a2252'
down_revision = '16f1fbcab42b'
branch_labels = None
depends_on = None


def upgrade():
    for prefix in ('', 'shadow_'):
        table_name = prefix + 'block_device_mapping'
        with op.batch_alter_table(table_name, schema=None) as batch_op:
            batch_op.add_column(
                sa.Column(
                    'encrypted',
                    sa.Boolean(),
                    nullable=True,
                )
            )
            batch_op.add_column(
                sa.Column(
                    'encryption_secret_uuid',
                    sa.String(length=36),
                    nullable=True,
                )
            )
            batch_op.add_column(
                sa.Column('encryption_format',
                    sa.String(length=128),
                    nullable=True,
                )
            )
            batch_op.add_column(
                sa.Column('encryption_options',
                    sa.String(length=4096),
                    nullable=True,
                )
            )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from oslo_messaging import exceptions as messaging_exceptions
from oslo_utils.fixture import uuidsentinel as uuids
from oslo_utils import timeutils

from nova.compute import instance_actions
from nova.compute import power_state
from nova.compute import task_states
from nova.compute import utils as compute_utils
from nova.compute import vm_states
from nova.conductor.tasks import cross_cell_migrate
from nova import context as nova_context
from nova import exception
from nova.network import model as network_model
from nova import objects
from nova.objects import base as obj_base
from nova.objects import fields
from nova.objects import instance as instance_obj
from nova import test
from nova.tests.unit.db.main import test_api as test_db_api
from nova.tests.unit import fake_block_device
from nova.tests.unit import fake_instance
from nova.tests.unit.objects import test_compute_node
from nova.tests.unit.objects import test_instance_device_metadata
from nova.tests.unit.objects import test_instance_numa
from nova.tests.unit.objects import test_instance_pci_requests
from nova.tests.unit.objects import test_keypair
from nova.tests.unit.objects import test_migration
from nova.tests.unit.objects import test_pci_device
from nova.tests.unit.objects import test_service
from nova.tests.unit.objects import test_vcpu_model


class ObjectComparatorMixin(test_db_api.ModelsObjectComparatorMixin):
    """Mixin class to aid in comparing two objects."""

    def _compare_objs(self, obj1, obj2, ignored_keys=None):
        # We can always ignore id since it is not deterministic when records
        # are copied over to the target cell database.
        if ignored_keys is None:
            ignored_keys = []
        if 'id' not in ignored_keys:
            ignored_keys.append('id')
        prim1 = obj1.obj_to_primitive()['nova_object.data']
        prim2 = obj2.obj_to_primitive()['nova_object.data']
        if isinstance(obj1, obj_base.ObjectListBase):
            self.assertEqual(len(obj1), len(obj2))
            prim1 = [o['nova_object.data'] for o in prim1['objects']]
            prim2 = [o['nova_object.data'] for o in prim2['objects']]
            self._assertEqualListsOfObjects(
                prim1, prim2, ignored_keys=ignored_keys)
        else:
            self._assertEqualObjects(prim1, prim2, ignored_keys=ignored_keys)


class TargetDBSetupTaskTestCase(test.TestCase, ObjectComparatorMixin):

    def setUp(self):
        super(TargetDBSetupTaskTestCase, self).setUp()
        cells = list(self.cell_mappings.values())
        self.source_cell = cells[0]
        self.target_cell = cells[1]
        # Pass is_admin=True because of the funky DB API
        # _check_instance_exists_in_project check when creating instance tags.
        self.source_context = nova_context.RequestContext(
            user_id='fake-user', project_id='fake-project', is_admin=True)
        self.target_context = self.source_context.elevated()  # copy source
        nova_context.set_target_cell(self.source_context, self.source_cell)
        nova_context.set_target_cell(self.target_context, self.target_cell)

    def _create_instance_data(self):
        """Creates an instance record and associated data like BDMs, VIFs,
        migrations, etc in the source cell and returns the Instance object.

        The idea is to create as many things from the
        Instance.INSTANCE_OPTIONAL_ATTRS list as possible.

        :returns: The created Instance and Migration objects
        """
        # Create the nova-compute services record first.
        fake_service = test_service._fake_service()
        fake_service.pop('version', None)  # version field is immutable
        fake_service.pop('id', None)  # cannot create with an id set
        service = objects.Service(self.source_context, **fake_service)
        service.create()
        # Create the compute node using the service.
        fake_compute_node = copy.copy(test_compute_node.fake_compute_node)
        fake_compute_node['host'] = service.host
        fake_compute_node['hypervisor_hostname'] = service.host
        fake_compute_node['stats'] = {}  # the object requires a dict
        fake_compute_node['service_id'] = service.id
        fake_compute_node.pop('id', None)  # cannot create with an id set
        compute_node = objects.ComputeNode(
            self.source_context, **fake_compute_node)
        compute_node.create()

        # Build an Instance object with basic fields set.
        updates = {
            'metadata': {'foo': 'bar'},
            'system_metadata': {'roles': ['member']},
            'host': compute_node.host,
            'node': compute_node.hypervisor_hostname
        }
        inst = fake_instance.fake_instance_obj(self.source_context, **updates)
        delattr(inst, 'id')  # cannot create an instance with an id set
        # Now we have to dirty all of the fields because fake_instance_obj
        # uses Instance._from_db_object to create the Instance object we have
        # but _from_db_object calls obj_reset_changes() which resets all of
        # the fields that were on the object, including the basic stuff like
        # the 'host' field, which means those fields don't get set in the DB.
        # TODO(mriedem): This should live in fake_instance_obj with a
        # make_creatable kwarg.
        for field in inst.obj_fields:
            if field in inst:
                setattr(inst, field, getattr(inst, field))
        # Make sure at least one expected basic field is dirty on the Instance.
        self.assertIn('host', inst.obj_what_changed())
        # Set the optional fields on the instance before creating it.
        inst.pci_requests = objects.InstancePCIRequests(requests=[
            objects.InstancePCIRequest(
                **test_instance_pci_requests.fake_pci_requests[0])])
        inst.numa_topology = objects.InstanceNUMATopology(
            cells=test_instance_numa.fake_obj_numa_topology.cells)
        inst.trusted_certs = objects.TrustedCerts(ids=[uuids.cert])
        inst.vcpu_model = test_vcpu_model.fake_vcpumodel
        inst.keypairs = objects.KeyPairList(objects=[
            objects.KeyPair(**test_keypair.fake_keypair)])
        inst.device_metadata = (
            test_instance_device_metadata.get_fake_obj_device_metadata(
                self.source_context))
        # FIXME(mriedem): db.instance_create does not handle tags
        inst.obj_reset_changes(['tags'])
        inst.create()

        bdm = {
            'instance_uuid': inst.uuid,
            'source_type': 'volume',
            'destination_type': 'volume',
            'volume_id': uuids.volume_id,
            'volume_size': 1,
            'device_name': '/dev/vda',
        }
        bdm = objects.BlockDeviceMapping(
            self.source_context,
            **fake_block_device.FakeDbBlockDeviceDict(bdm_dict=bdm))
        delattr(bdm, 'id')  # cannot create a bdm with an id set
        bdm.obj_reset_changes(['id'])
        bdm.create()

        vif = objects.VirtualInterface(
            self.source_context, address='de:ad:be:ef:ca:fe', uuid=uuids.port,
            instance_uuid=inst.uuid)
        vif.create()

        info_cache = objects.InstanceInfoCache().new(
            self.source_context, inst.uuid)
        info_cache.network_info = network_model.NetworkInfo([
                network_model.VIF(id=vif.uuid, address=vif.address)])
        info_cache.save(update_cells=False)

        objects.TagList.create(self.source_context, inst.uuid, ['test'])

        try:
            raise test.TestingException('test-fault')
        except test.TestingException as fault:
            compute_utils.add_instance_fault_from_exc(
                self.source_context, inst, fault)

        objects.InstanceAction().action_start(
            self.source_context, inst.uuid, 'resize', want_result=False)
        objects.InstanceActionEvent().event_start(
            self.source_context, inst.uuid, 'migrate_server',
            want_result=False)

        # Create a fake migration for the cross-cell resize operation.
        migration = objects.Migration(
            self.source_context,
            **test_migration.fake_db_migration(
                instance_uuid=inst.uuid, cross_cell_move=True,
                migration_type='resize'))
        delattr(migration, 'id')  # cannot create a migration with an id set
        migration.obj_reset_changes(['id'])
        migration.create()

        # Create an old non-resize migration to make sure it is copied to the
        # target cell database properly.
        old_migration = objects.Migration(
            self.source_context,
            **test_migration.fake_db_migration(
                instance_uuid=inst.uuid, migration_type='live-migration',
                status='completed', uuid=uuids.old_migration))
        delattr(old_migration, 'id')  # cannot create a migration with an id
        old_migration.obj_reset_changes(['id'])
        old_migration.create()

        fake_pci_device = copy.copy(test_pci_device.fake_db_dev)
        fake_pci_device['extra_info'] = {}  # the object requires a dict
        fake_pci_device['compute_node_id'] = compute_node.id
        pci_device = objects.PciDevice.create(
            self.source_context, fake_pci_device)
        pci_device.allocate(inst)  # sets the status and instance_uuid fields
        pci_device.save()

        # Return a fresh copy of the instance from the DB with as many joined
        # fields loaded as possible.
        expected_attrs = copy.copy(instance_obj.INSTANCE_OPTIONAL_ATTRS)
        # Cannot load fault from get_by_uuid.
        expected_attrs.remove('fault')
        inst = objects.Instance.get_by_uuid(
            self.source_context, inst.uuid, expected_attrs=expected_attrs)
        return inst, migration

    def test_execute_and_rollback(self):
        """Happy path test which creates an instance with related records
        in a source cell and then executes TargetDBSetupTask to create those
        same records in a target cell. Runs rollback to make sure the target
        cell instance is deleted.
        """
        source_cell_instance, migration = self._create_instance_data()
        instance_uuid = source_cell_instance.uuid

        task = cross_cell_migrate.TargetDBSetupTask(
            self.source_context, source_cell_instance, migration,
            self.target_context)
        target_cell_instance = task.execute()[0]

        # The instance in the target cell should be hidden.
        self.assertTrue(target_cell_instance.hidden,
                        'Target cell instance should be hidden')
        # Assert that the various records created in _create_instance_data are
        # found in the target cell database. We ignore 'hidden' because the
        # values are explicitly different between source and target DB. The
        # pci_devices and services fields are not set on the target instance
        # during TargetDBSetupTask.execute so we ignore those here and verify
        # them below. tags are also special in that we have to lazy-load them
        # on target_cell_instance so we check those explicitly below as well.
        ignored_keys = ['hidden', 'pci_devices', 'services', 'tags']
        self._compare_objs(source_cell_instance, target_cell_instance,
                           ignored_keys=ignored_keys)

        # Explicitly compare flavor fields to make sure they are created and
        # loaded properly.
        for flavor_field in ('old_', 'new_', ''):
            source_field = getattr(
                source_cell_instance, flavor_field + 'flavor')
            target_field = getattr(
                target_cell_instance, flavor_field + 'flavor')
            # old/new may not be set
            if source_field is None or target_field is None:
                self.assertIsNone(source_field)
                self.assertIsNone(target_field)
            else:
                self._compare_objs(source_field, target_field)

        # Compare PCI requests
        self.assertIsNotNone(target_cell_instance.pci_requests)
        self._compare_objs(source_cell_instance.pci_requests,
                           target_cell_instance.pci_requests)

        # Compare requested instance NUMA topology
        self.assertIsNotNone(target_cell_instance.numa_topology)
        self._compare_objs(source_cell_instance.numa_topology,
                           target_cell_instance.numa_topology)

        # Compare trusted certs
        self.assertIsNotNone(target_cell_instance.trusted_certs)
        self._compare_objs(source_cell_instance.trusted_certs,
                           target_cell_instance.trusted_certs)

        # Compare vcpu_model
        self.assertIsNotNone(target_cell_instance.vcpu_model)
        self._compare_objs(source_cell_instance.vcpu_model,
                           target_cell_instance.vcpu_model)

        # Compare keypairs
        self.assertEqual(1, len(target_cell_instance.keypairs))
        self._compare_objs(source_cell_instance.keypairs,
                           target_cell_instance.keypairs)

        # Compare device_metadata
        self.assertIsNotNone(target_cell_instance.device_metadata)
        self._compare_objs(source_cell_instance.device_metadata,
                           target_cell_instance.device_metadata)

        # Compare BDMs
        target_bdms = target_cell_instance.get_bdms()
        self.assertEqual(1, len(target_bdms))
        self._compare_objs(source_cell_instance.get_bdms(), target_bdms)
        self.assertEqual(source_cell_instance.uuid,
                         target_bdms[0].instance_uuid)

        # Compare VIFs
        source_vifs = objects.VirtualInterfaceList.get_by_instance_uuid(
            self.source_context, instance_uuid)
        target_vifs = objects.VirtualInterfaceList.get_by_instance_uuid(
            self.target_context, instance_uuid)
        self.assertEqual(1, len(target_vifs))
        self._compare_objs(source_vifs, target_vifs)

        # Compare info cache (there should be a single vif in the target)
        self.assertEqual(1, len(target_cell_instance.info_cache.network_info))
        self.assertEqual(target_vifs[0].uuid,
                         target_cell_instance.info_cache.network_info[0]['id'])
        self._compare_objs(source_cell_instance.info_cache,
                           target_cell_instance.info_cache)

        # Compare tags
        self.assertEqual(1, len(target_cell_instance.tags))
        self._compare_objs(source_cell_instance.tags,
                           target_cell_instance.tags)

        # Assert that the fault from the source is not in the target.
        self.assertIsNone(target_cell_instance.fault)

        # Compare instance actions and events
        source_actions = objects.InstanceActionList.get_by_instance_uuid(
            self.source_context, instance_uuid)
        target_actions = objects.InstanceActionList.get_by_instance_uuid(
            self.target_context, instance_uuid)
        self._compare_objs(source_actions, target_actions)

        # The InstanceActionEvent.action_id is per-cell DB so we need to get
        # the events per action and compare them but ignore the action_id.
        source_events = objects.InstanceActionEventList.get_by_action(
            self.source_context, source_actions[0].id)
        target_events = objects.InstanceActionEventList.get_by_action(
            self.target_context, target_actions[0].id)
        self._compare_objs(source_events, target_events,
                           ignored_keys=['action_id'])

        # Compare migrations
        filters = {'instance_uuid': instance_uuid}
        source_migrations = objects.MigrationList.get_by_filters(
            self.source_context, filters)
        target_migrations = objects.MigrationList.get_by_filters(
            self.target_context, filters)
        # There should be two migrations in the target cell.
        self.assertEqual(2, len(target_migrations))
        self._compare_objs(source_migrations, target_migrations)
        # One should be a live-migration type (make sure Migration._from-db_obj
        # did not set the migration_type for us).
        migration_types = [mig.migration_type for mig in target_migrations]
        self.assertIn('resize', migration_types)
        self.assertIn('live-migration', migration_types)

        # pci_devices and services should not have been copied over since they
        # are specific to the compute node in the source cell database
        for field in ('pci_devices', 'services'):
            source_value = getattr(source_cell_instance, field)
            self.assertEqual(
                1, len(source_value),
                'Unexpected number of %s in source cell instance' % field)
            target_value = getattr(target_cell_instance, field)
            self.assertEqual(
                0, len(target_value),
                'Unexpected number of %s in target cell instance' % field)

        # Rollback the task and assert the instance and its related data are
        # gone from the target cell database. Use a modified context to make
        # sure the instance was hard-deleted.
        task.rollback(test.TestingException('error'))
        read_deleted_ctxt = self.target_context.elevated(read_deleted='yes')
        self.assertRaises(exception.InstanceNotFound,
                          objects.Instance.get_by_uuid,
                          read_deleted_ctxt, target_cell_instance.uuid)


class CrossCellMigrationTaskTestCase(test.NoDBTestCase):

    def setUp(self):
        super(CrossCellMigrationTaskTestCase, self).setUp()
        source_context = nova_context.get_context()
        host_selection = objects.Selection(
            service_host='target.host.com', cell_uuid=uuids.cell_uuid)
        migration = objects.Migration(
            id=1, cross_cell_move=False, source_compute='source.host.com')
        instance = objects.Instance()
        self.task = cross_cell_migrate.CrossCellMigrationTask(
            source_context,
            instance,
            objects.Flavor(),
            mock.sentinel.request_spec,
            migration,
            mock.sentinel.compute_rpcapi,
            host_selection,
            mock.sentinel.alternate_hosts)

    def test_execute_and_rollback(self):
        """Basic test to just hit execute and rollback."""
        # Mock out the things that execute calls
        with test.nested(
            mock.patch.object(self.task.source_migration, 'save'),
            mock.patch.object(self.task, '_perform_external_api_checks'),
            mock.patch.object(self.task, '_setup_target_cell_db'),
            mock.patch.object(self.task, '_prep_resize_at_dest'),
            mock.patch.object(self.task, '_prep_resize_at_source'),
            mock.patch.object(self.task, '_finish_resize_at_dest'),
        ) as (
            mock_migration_save, mock_perform_external_api_checks,
            mock_setup_target_cell_db, mock_prep_resize_at_dest,
            mock_prep_resize_at_source, mock_finish_resize_at_dest,
        ):
            mock_setup_target_cell_db.return_value = (
                mock.sentinel.target_cell_migration,
                mock.sentinel.target_cell_mapping)
            self.task.execute()
        # Assert the calls
        self.assertTrue(self.task.source_migration.cross_cell_move,
                        'Migration.cross_cell_move should be True.')
        mock_migration_save.assert_called_once_with()
        mock_perform_external_api_checks.assert_called_once_with()
        mock_setup_target_cell_db.assert_called_once_with()
        mock_prep_resize_at_dest.assert_called_once_with(
            mock.sentinel.target_cell_migration)
        mock_prep_resize_at_source.assert_called_once_with()
        mock_finish_resize_at_dest.assert_called_once_with(
            mock_prep_resize_at_dest.return_value,
            mock.sentinel.target_cell_mapping,
            mock_prep_resize_at_source.return_value)
        # Now rollback the completed sub-tasks
        self.task.rollback(test.TestingException('error'))

    def test_perform_external_api_checks_ok(self):
        """Tests the happy path scenario where neutron APIs are new enough for
        what we need.
        """
        with mock.patch.object(
                self.task.network_api, 'has_port_binding_extension',
                return_value=True) as mock_neutron_check:
            self.task._perform_external_api_checks()
        mock_neutron_check.assert_called_once_with(self.task.context)

    def test_perform_external_api_checks_old_neutron(self):
        """Tests the case that neutron API is old."""
        with mock.patch.object(
                self.task.network_api, 'has_port_binding_extension',
                return_value=False):
            ex = self.assertRaises(exception.MigrationPreCheckError,
                                   self.task._perform_external_api_checks)
            self.assertIn('Required networking service API extension', str(ex))

    @mock.patch('nova.conductor.tasks.cross_cell_migrate.LOG.exception')
    def test_rollback_idempotent(self, mock_log_exception):
        """Tests that the rollback routine hits all completed tasks even if
        one or more of them fail their own rollback routine.
        """
        # Mock out some completed tasks
        for x in range(3):
            task = mock.Mock()
            # The 2nd task will fail its rollback.
            if x == 1:
                task.rollback.side_effect = test.TestingException('sub-task')
            self.task._completed_tasks[str(x)] = task
        # Run execute but mock _execute to fail somehow.
        error = test.TestingException('main task')
        with mock.patch.object(self.task, '_execute', side_effect=error):
            # The TestingException from the main task should be raised.
            ex = self.assertRaises(test.TestingException, self.task.execute)
            self.assertEqual('main task', str(ex))
        # And all three sub-task rollbacks should have been called.
        for subtask in self.task._completed_tasks.values():
            subtask.rollback.assert_called_once_with(error)
        # The 2nd task rollback should have raised and been logged.
        mock_log_exception.assert_called_once()
        self.assertEqual('1', mock_log_exception.call_args[0][1])

    @mock.patch('nova.objects.CellMapping.get_by_uuid')
    @mock.patch('nova.context.set_target_cell')
    @mock.patch.object(cross_cell_migrate.TargetDBSetupTask, 'execute')
    def test_setup_target_cell_db(self, mock_target_db_set_task_execute,
                                  mock_set_target_cell, mock_get_cell_mapping):
        """Tests setting up and executing TargetDBSetupTask"""
        mock_target_db_set_task_execute.return_value = (
            mock.sentinel.target_cell_instance,
            mock.sentinel.target_cell_migration)
        result = self.task._setup_target_cell_db()
        mock_target_db_set_task_execute.assert_called_once_with()
        mock_get_cell_mapping.assert_called_once_with(
            self.task.context, self.task.host_selection.cell_uuid)
        # The target_cell_context should be set on the main task but as a copy
        # of the source context.
        self.assertIsNotNone(self.task._target_cell_context)
        self.assertIsNot(self.task._target_cell_context, self.task.context)
        # The target cell context should have been targeted to the target
        # cell mapping.
        mock_set_target_cell.assert_called_once_with(
            self.task._target_cell_context, mock_get_cell_mapping.return_value)
        # The resulting migration record from TargetDBSetupTask should have
        # been returned along with the target cell mapping.
        self.assertIs(result[0], mock.sentinel.target_cell_migration)
        self.assertIs(result[1], mock_get_cell_mapping.return_value)
        # The target_cell_instance should be set on the main task.
        self.assertIsNotNone(self.task._target_cell_instance)
        self.assertIs(self.task._target_cell_instance,
                      mock.sentinel.target_cell_instance)
        # And the completed task should have been recorded for rollbacks.
        self.assertIn('TargetDBSetupTask', self.task._completed_tasks)
        self.assertIsInstance(self.task._completed_tasks['TargetDBSetupTask'],
                              cross_cell_migrate.TargetDBSetupTask)

    @mock.patch.object(cross_cell_migrate.PrepResizeAtDestTask, 'execute')
    @mock.patch('nova.availability_zones.get_host_availability_zone',
                return_value='cell2-az1')
    def test_prep_resize_at_dest(self, mock_get_az, mock_task_execute):
        """Tests setting up and executing PrepResizeAtDestTask"""
        # _setup_target_cell_db set the _target_cell_context and
        # _target_cell_instance variables so fake those out here
        self.task._target_cell_context = mock.sentinel.target_cell_context
        target_inst = objects.Instance(
            vm_state=vm_states.ACTIVE, system_metadata={})
        self.task._target_cell_instance = target_inst
        target_cell_migration = objects.Migration(
            # use unique ids for comparisons
            id=self.task.source_migration.id + 1)
        self.assertNotIn('migration_context', self.task.instance)
        mock_task_execute.return_value = objects.MigrationContext(
            migration_id=target_cell_migration.id)

        with test.nested(
            mock.patch.object(self.task,
                              '_update_migration_from_dest_after_claim'),
            mock.patch.object(self.task.instance, 'save'),
            mock.patch.object(target_inst, 'save')
        ) as (
            _upd_mig, source_inst_save, target_inst_save
        ):
            retval = self.task._prep_resize_at_dest(target_cell_migration)

        self.assertIs(retval, _upd_mig.return_value)
        mock_task_execute.assert_called_once_with()
        mock_get_az.assert_called_once_with(
            self.task.context, self.task.host_selection.service_host)
        self.assertIn('PrepResizeAtDestTask', self.task._completed_tasks)
        self.assertIsInstance(
            self.task._completed_tasks['PrepResizeAtDestTask'],
            cross_cell_migrate.PrepResizeAtDestTask)
        # The new_flavor should be set on the target cell instance along with
        # the AZ and old_vm_state.
        self.assertIs(target_inst.new_flavor, self.task.flavor)
        self.assertEqual(vm_states.ACTIVE,
                         target_inst.system_metadata['old_vm_state'])
        self.assertEqual(mock_get_az.return_value,
                         target_inst.availability_zone)
        # A clone of the MigrationContext returned from execute() should be
        # stored on the source instance with the internal context targeted
        # at the source cell context and the migration_id updated.
        self.assertIsNotNone(self.task.instance.migration_context)
        self.assertEqual(self.task.source_migration.id,
                         self.task.instance.migration_context.migration_id)
        source_inst_save.assert_called_once_with()
        _upd_mig.assert_called_once_with(target_cell_migration)

    @mock.patch('nova.objects.Migration.get_by_uuid')
    def test_update_migration_from_dest_after_claim(self, get_by_uuid):
        """Tests the _update_migration_from_dest_after_claim method."""
        self.task._target_cell_context = mock.sentinel.target_cell_context
        target_cell_migration = objects.Migration(
            uuid=uuids.migration, cross_cell_move=True,
            dest_compute='dest-compute', dest_node='dest-node',
            dest_host='192.168.159.176')
        get_by_uuid.return_value = target_cell_migration.obj_clone()
        with mock.patch.object(self.task.source_migration, 'save') as save:
            retval = self.task._update_migration_from_dest_after_claim(
                target_cell_migration)
        # The returned target cell migration should be the one we pulled from
        # the target cell database.
        self.assertIs(retval, get_by_uuid.return_value)
        get_by_uuid.assert_called_once_with(
            self.task._target_cell_context, target_cell_migration.uuid)
        # The source cell migration on the task should have been updated.
        source_cell_migration = self.task.source_migration
        self.assertEqual('dest-compute', source_cell_migration.dest_compute)
        self.assertEqual('dest-node', source_cell_migration.dest_node)
        self.assertEqual('192.168.159.176', source_cell_migration.dest_host)
        save.assert_called_once_with()

    @mock.patch.object(cross_cell_migrate.PrepResizeAtSourceTask, 'execute')
    def test_prep_resize_at_source(self, mock_task_execute):
        """Tests setting up and executing PrepResizeAtSourceTask"""
        snapshot_id = self.task._prep_resize_at_source()
        self.assertIs(snapshot_id, mock_task_execute.return_value)
        self.assertIn('PrepResizeAtSourceTask', self.task._completed_tasks)
        self.assertIsInstance(
            self.task._completed_tasks['PrepResizeAtSourceTask'],
            cross_cell_migrate.PrepResizeAtSourceTask)

    @mock.patch.object(cross_cell_migrate.FinishResizeAtDestTask, 'execute')
    def test_finish_resize_at_dest(self, mock_task_execute):
        """Tests setting up and executing FinishResizeAtDestTask"""
        target_cell_migration = objects.Migration()
        target_cell_mapping = objects.CellMapping()
        self.task._finish_resize_at_dest(
            target_cell_migration, target_cell_mapping, uuids.snapshot_id)
        mock_task_execute.assert_called_once_with()
        self.assertIn('FinishResizeAtDestTask', self.task._completed_tasks)
        self.assertIsInstance(
            self.task._completed_tasks['FinishResizeAtDestTask'],
            cross_cell_migrate.FinishResizeAtDestTask)


class PrepResizeAtDestTaskTestCase(test.NoDBTestCase):

    def setUp(self):
        super(PrepResizeAtDestTaskTestCase, self).setUp()
        host_selection = objects.Selection(
            service_host='fake-host', nodename='fake-host',
            limits=objects.SchedulerLimits())
        self.task = cross_cell_migrate.PrepResizeAtDestTask(
            nova_context.get_context(),
            objects.Instance(uuid=uuids.instance),
            objects.Flavor(),
            objects.Migration(),
            objects.RequestSpec(),
            compute_rpcapi=mock.Mock(),
            host_selection=host_selection,
            network_api=mock.Mock(),
            volume_api=mock.Mock())

    def test_create_port_bindings(self):
        """Happy path test for creating port bindings"""
        with mock.patch.object(
                self.task.network_api, 'bind_ports_to_host') as mock_bind:
            self.task._create_port_bindings()
        self.assertIs(self.task._bindings_by_port_id, mock_bind.return_value)
        mock_bind.assert_called_once_with(
            self.task.context, self.task.instance,
            self.task.host_selection.service_host)

    def test_create_port_bindings_port_binding_failed(self):
        """Tests that bind_ports_to_host raises PortBindingFailed which
        results in a MigrationPreCheckError.
        """
        with mock.patch.object(
                self.task.network_api, 'bind_ports_to_host',
                side_effect=exception.PortBindingFailed(
                    port_id=uuids.port_id)) as mock_bind:
            self.assertRaises(exception.MigrationPreCheckError,
                              self.task._create_port_bindings)
        self.assertEqual({}, self.task._bindings_by_port_id)
        mock_bind.assert_called_once_with(
            self.task.context, self.task.instance,
            self.task.host_selection.service_host)

    @mock.patch('nova.objects.BlockDeviceMapping.save')
    def test_create_volume_attachments(self, mock_bdm_save):
        """Happy path test for creating volume attachments"""
        # Two BDMs: one as a local image and one as an attached data volume;
        # only the volume BDM should be processed and returned.
        bdms = objects.BlockDeviceMappingList(objects=[
            objects.BlockDeviceMapping(
                source_type='image', destination_type='local'),
            objects.BlockDeviceMapping(
                source_type='volume', destination_type='volume',
                volume_id=uuids.volume_id,
                instance_uuid=self.task.instance.uuid)])
        with test.nested(
            mock.patch.object(
                self.task.instance, 'get_bdms', return_value=bdms),
            mock.patch.object(
                self.task.volume_api, 'attachment_create',
                return_value={'id': uuids.attachment_id}),
        ) as (
            mock_get_bdms, mock_attachment_create
        ):
            volume_bdms = self.task._create_volume_attachments()

        mock_attachment_create.assert_called_once_with(
            self.task.context, uuids.volume_id, self.task.instance.uuid)
        # The created attachment ID should be saved for rollbacks.
        self.assertEqual(1, len(self.task._created_volume_attachment_ids))
        self.assertEqual(
            uuids.attachment_id, self.task._created_volume_attachment_ids[0])
        # Only the volume BDM should have been processed and returned.
        self.assertEqual(1, len(volume_bdms))
        self.assertIs(bdms[1], volume_bdms[0])
        # The volume BDM attachment_id should have been updated.
        self.assertEqual(uuids.attachment_id, volume_bdms[0].attachment_id)

    def test_execute(self):
        """Happy path for executing the task"""

        def fake_create_port_bindings():
            self.task._bindings_by_port_id = mock.sentinel.bindings

        with test.nested(
            mock.patch.object(self.task, '_create_port_bindings',
                              side_effect=fake_create_port_bindings),
            mock.patch.object(self.task, '_create_volume_attachments'),
            mock.patch.object(
                self.task.compute_rpcapi, 'prep_snapshot_based_resize_at_dest')
        ) as (
            _create_port_bindings, _create_volume_attachments,
            prep_snapshot_based_resize_at_dest
        ):
            # Execute the task. The return value should be the MigrationContext
            # returned from prep_snapshot_based_resize_at_dest.
            self.assertEqual(
                prep_snapshot_based_resize_at_dest.return_value,
                self.task.execute())

        _create_port_bindings.assert_called_once_with()
        _create_volume_attachments.assert_called_once_with()
        prep_snapshot_based_resize_at_dest.assert_called_once_with(
            self.task.context, self.task.instance, self.task.flavor,
            self.task.host_selection.nodename, self.task.target_migration,
            self.task.host_selection.limits, self.task.request_spec,
            self.task.host_selection.service_host)

    def test_execute_messaging_timeout(self):
        """Tests the case that prep_snapshot_based_resize_at_dest raises
        MessagingTimeout which results in a MigrationPreCheckError.
        """
        with test.nested(
            mock.patch.object(self.task, '_create_port_bindings'),
            mock.patch.object(self.task, '_create_volume_attachments'),
            mock.patch.object(
                self.task.compute_rpcapi, 'prep_snapshot_based_resize_at_dest',
                side_effect=messaging_exceptions.MessagingTimeout)
        ) as (
            _create_port_bindings, _create_volume_attachments,
            prep_snapshot_based_resize_at_dest
        ):
            ex = self.assertRaises(
                exception.MigrationPreCheckError, self.task.execute)
            self.assertIn(
                'RPC timeout while checking if we can cross-cell migrate to '
                'host: fake-host', str(ex))

        _create_port_bindings.assert_called_once_with()
        _create_volume_attachments.assert_called_once_with()
        prep_snapshot_based_resize_at_dest.assert_called_once_with(
            self.task.context, self.task.instance, self.task.flavor,
            self.task.host_selection.nodename, self.task.target_migration,
            self.task.host_selection.limits, self.task.request_spec,
            self.task.host_selection.service_host)

    @mock.patch('nova.conductor.tasks.cross_cell_migrate.LOG.exception')
    def test_rollback(self, mock_log_exception):
        """Tests rollback to make sure it idempotently handles cleaning up
        port bindings and volume attachments even if one in the set fails for
        each.
        """
        # Make sure we have two port bindings and two volume attachments
        # because we are going to make the first of each fail and we want to
        # make sure we still try to delete the other.
        self.task._bindings_by_port_id = {
            uuids.port_id1: mock.sentinel.binding1,
            uuids.port_id2: mock.sentinel.binding2
        }
        self.task._created_volume_attachment_ids = [
            uuids.attachment_id1, uuids.attachment_id2
        ]
        with test.nested(
            mock.patch.object(
                self.task.network_api, 'delete_port_binding',
                # First call fails, second is OK.
                side_effect=(exception.PortBindingDeletionFailed, None)),
            mock.patch.object(
                self.task.volume_api, 'attachment_delete',
                # First call fails, second is OK.
                side_effect=(exception.CinderConnectionFailed, None)),
        ) as (
            delete_port_binding, attachment_delete
        ):
            self.task.rollback(test.TestingException('error'))
        # Should have called both delete methods twice in any order.
        host = self.task.host_selection.service_host
        delete_port_binding.assert_has_calls([
            mock.call(self.task.context, port_id, host)
            for port_id in self.task._bindings_by_port_id],
            any_order=True)
        attachment_delete.assert_has_calls([
            mock.call(self.task.context, attachment_id)
            for attachment_id in self.task._created_volume_attachment_ids],
            any_order=True)
        # Should have logged both exceptions.
        self.assertEqual(2, mock_log_exception.call_count)


class PrepResizeAtSourceTaskTestCase(test.NoDBTestCase):

    def setUp(self):
        super(PrepResizeAtSourceTaskTestCase, self).setUp()
        self.task = cross_cell_migrate.PrepResizeAtSourceTask(
            nova_context.get_context(),
            objects.Instance(
                uuid=uuids.instance,
                vm_state=vm_states.ACTIVE,
                display_name='fake-server',
                system_metadata={},
                host='source.host.com',
                flavor=objects.Flavor(),
            ),
            objects.Flavor(),
            objects.Migration(),
            objects.RequestSpec(),
            compute_rpcapi=mock.Mock(),
            image_api=mock.Mock())

    @mock.patch('nova.compute.utils.create_image')
    @mock.patch('nova.objects.Instance.save')
    def test_execute_volume_backed(self, instance_save, create_image):
        """Tests execution with a volume-backed server so no snapshot image
        is created.
        """
        self.task.request_spec.is_bfv = True
        # No image should be created so no image is returned.
        self.assertIsNone(self.task.execute())
        self.assertIsNone(self.task._image_id)
        create_image.assert_not_called()
        self.task.compute_rpcapi.prep_snapshot_based_resize_at_source.\
            assert_called_once_with(
                self.task.context, self.task.instance, self.task.migration,
                snapshot_id=None)
        # The instance should have been updated.
        instance_save.assert_called_once_with(
            expected_task_state=task_states.RESIZE_PREP)
        self.assertIs(self.task.instance.old_flavor, self.task.instance.flavor)
        self.assertIs(self.task.instance.new_flavor, self.task.flavor)
        self.assertEqual(
            task_states.RESIZE_MIGRATING, self.task.instance.task_state)
        self.assertEqual(self.task.instance.vm_state,
                         self.task.instance.system_metadata['old_vm_state'])

    @mock.patch('nova.compute.utils.create_image',
                return_value={'id': uuids.snapshot_id})
    @mock.patch('nova.objects.Instance.save')
    def test_execute_image_backed(self, instance_save, create_image):
        """Tests execution with an image-backed server so a snapshot image
        is created.
        """
        self.task.request_spec.is_bfv = False
        self.task.instance.image_ref = uuids.old_image_ref
        # An image should be created so an image ID is returned.
        self.assertEqual(uuids.snapshot_id, self.task.execute())
        self.assertEqual(uuids.snapshot_id, self.task._image_id)
        create_image.assert_called_once_with(
            self.task.context, self.task.instance, 'fake-server-resize-temp',
            'snapshot', self.task.image_api)
        self.task.compute_rpcapi.prep_snapshot_based_resize_at_source.\
            assert_called_once_with(
                self.task.context, self.task.instance, self.task.migration,
                snapshot_id=uuids.snapshot_id)
        # The instance should have been updated.
        instance_save.assert_called_once_with(
            expected_task_state=task_states.RESIZE_PREP)
        self.assertIs(self.task.instance.old_flavor, self.task.instance.flavor)
        self.assertIs(self.task.instance.new_flavor, self.task.flavor)
        self.assertEqual(
            task_states.RESIZE_MIGRATING, self.task.instance.task_state)
        self.assertEqual(self.task.instance.vm_state,
                         self.task.instance.system_metadata['old_vm_state'])

    @mock.patch('nova.compute.utils.delete_image')
    def test_rollback(self, delete_image):
        """Tests rollback when there is an image and when there is not."""
        # First test when there is no image_id so we do not try to delete it.
        self.task.rollback(test.TestingException('error'))
        delete_image.assert_not_called()
        # Now set an image and we should try to delete it.
        self.task._image_id = uuids.image_id
        self.task.rollback(test.TestingException('error'))
        delete_image.assert_called_once_with(
            self.task.context, self.task.instance, self.task.image_api,
            self.task._image_id)


class FinishResizeAtDestTaskTestCase(test.TestCase):
    """Tests for FinishResizeAtDestTask which rely on a database"""

    def _create_instance(self, ctxt, create_instance_mapping=False, **updates):
        """Create a fake instance with the given cell-targeted context

        :param ctxt: Cell-targeted RequestContext
        :param create_instance_mapping: If True, create an InstanceMapping
            for the instance pointed at the cell in which the ctxt is targeted,
            otherwise no InstanceMapping is created.
        :param updates: Additional fields to set on the Instance object.
        :returns: Instance object that was created.
        """
        inst = fake_instance.fake_instance_obj(ctxt, **updates)
        delattr(inst, 'id')  # make it creatable
        # Now we have to dirty all of the fields because fake_instance_obj
        # uses Instance._from_db_object to create the Instance object we have
        # but _from_db_object calls obj_reset_changes() which resets all of
        # the fields that were on the object, including the basic stuff like
        # the 'host' field, which means those fields don't get set in the DB.
        # TODO(mriedem): This should live in fake_instance_obj
        for field in inst.obj_fields:
            if field in inst:
                setattr(inst, field, getattr(inst, field))
        # FIXME(mriedem): db.instance_create does not handle tags
        inst.obj_reset_changes(['tags'])
        inst.create()

        if create_instance_mapping:
            # Find the cell mapping from the context.
            self.assertIsNotNone(ctxt.cell_uuid,
                                 'ctxt must be targeted to a cell.')
            for cell in self.cell_mappings.values():
                if cell.uuid == ctxt.cell_uuid:
                    break
            else:
                raise Exception('Unable to find CellMapping with UUID %s' %
                                ctxt.cell_uuid)

            mapping = objects.InstanceMapping(
                ctxt, instance_uuid=inst.uuid,
                project_id=inst.project_id, cell_mapping=cell)
            mapping.create()

        return inst

    def setUp(self):
        super(FinishResizeAtDestTaskTestCase, self).setUp()
        cells = list(self.cell_mappings.values())
        source_cell = cells[0]
        target_cell = cells[1]
        self.source_context = nova_context.RequestContext(
            user_id='fake-user', project_id='fake-project', is_admin=True)
        self.target_context = self.source_context.elevated()  # copy source
        nova_context.set_target_cell(self.source_context, source_cell)
        nova_context.set_target_cell(self.target_context, target_cell)

        # Create the source cell instance.
        source_instance = self._create_instance(
            self.source_context, create_instance_mapping=True,
            hidden=False)
        # Create the instance action record in the source cell which is needed
        # by the EventReporter.
        objects.InstanceAction.action_start(
            self.source_context, source_instance.uuid,
            instance_actions.RESIZE, want_result=False)
        # Create the target cell instance which would normally be a clone of
        # the source cell instance but the only thing these tests care about
        # is that the UUID matches. The target cell instance is also hidden.
        target_instance = self._create_instance(
            self.target_context, hidden=True, uuid=source_instance.uuid)
        target_migration = objects.Migration(dest_compute='target.host.com')
        self.task = cross_cell_migrate.FinishResizeAtDestTask(
            self.target_context, target_instance, target_migration,
            source_instance, compute_rpcapi=mock.Mock(),
            target_cell_mapping=target_cell, snapshot_id=uuids.snapshot_id,
            request_spec=objects.RequestSpec())

    def test_execute(self):
        """Tests the happy path scenario for the task execution."""
        with test.nested(
            mock.patch.object(
                self.task.compute_rpcapi,
                'finish_snapshot_based_resize_at_dest'),
            mock.patch.object(self.task.instance, 'refresh')
        ) as (
            finish_resize, refresh
        ):
            self.task.execute()
        # _finish_snapshot_based_resize_at_dest will set the instance
        # task_state to resize_migrated, save the change, and call the
        # finish_snapshot_based_resize_at_dest method.
        target_instance = self.task.instance
        self.assertEqual(task_states.RESIZE_MIGRATED,
                         self.task.instance.task_state)
        finish_resize.assert_called_once_with(
            self.task.context, target_instance, self.task.migration,
            self.task.snapshot_id, self.task.request_spec)
        refresh.assert_called_once_with()
        # _update_instance_mapping will swap the hidden fields and update
        # the instance mapping to point at the target cell.
        self.assertFalse(target_instance.hidden,
                         'Target cell instance should not be hidden')
        source_instance = self.task.source_cell_instance
        source_instance.refresh()
        self.assertTrue(source_instance.hidden,
                        'Source cell instance should be hidden')
        mapping = objects.InstanceMapping.get_by_instance_uuid(
            self.task.context, target_instance.uuid)
        self.assertEqual(self.target_context.cell_uuid,
                         mapping.cell_mapping.uuid)

    @mock.patch('nova.objects.InstanceMapping.save')
    def test_finish_snapshot_based_resize_at_dest_fails(self, mock_im_save):
        """Tests when the finish_snapshot_based_resize_at_dest compute method
        raises an error.
        """
        with test.nested(
            mock.patch.object(self.task.compute_rpcapi,
                              'finish_snapshot_based_resize_at_dest',
                              side_effect=test.TestingException('oops')),
            mock.patch.object(self.task, '_copy_latest_fault'),
        ) as (
            finish_resize, copy_fault
        ):
            self.assertRaises(test.TestingException,
                              self.task._finish_snapshot_based_resize_at_dest)
        # The source cell instance should be in error state.
        source_instance = self.task.source_cell_instance
        source_instance.refresh()
        self.assertEqual(vm_states.ERROR, source_instance.vm_state)
        self.assertIsNone(source_instance.task_state)
        # And the latest fault and instance action event should have been
        # copied from the target cell DB to the source cell DB.
        copy_fault.assert_called_once_with(self.source_context)
        # Assert the event was recorded in the source cell DB.
        event_name = 'compute_finish_snapshot_based_resize_at_dest'
        action = objects.InstanceAction.get_by_request_id(
            source_instance._context, source_instance.uuid,
            source_instance._context.request_id)
        self.assertIsNotNone(action, 'InstanceAction not found.')
        events = objects.InstanceActionEventList.get_by_action(
            source_instance._context, action.id)
        self.assertEqual(1, len(events), events)
        self.assertEqual(event_name, events[0].event)
        self.assertEqual('Error', events[0].result)
        self.assertIn('_finish_snapshot_based_resize_at_dest',
                      events[0].traceback)
        self.assertEqual(self.task.migration.dest_compute, events[0].host)
        # Assert the instance mapping was never updated.
        mock_im_save.assert_not_called()

    def test_copy_latest_fault(self):
        """Tests _copy_latest_fault working as expected"""
        # Inject a fault in the target cell database.
        try:
            raise test.TestingException('test-fault')
        except test.TestingException as fault:
            compute_utils.add_instance_fault_from_exc(
                self.target_context, self.task.instance, fault)
        self.task._copy_latest_fault(self.source_context)
        # Now make sure that fault shows up in the source cell DB (it will
        # get lazy-loaded here).
        fault = self.task.source_cell_instance.fault
        self.assertIsNotNone(fault, 'Fault not copied to source cell DB')
        # And it's the fault we expect.
        self.assertEqual('TestingException', fault.message)

    @mock.patch('nova.conductor.tasks.cross_cell_migrate.LOG.exception')
    def test_copy_latest_fault_error(self, mock_log):
        """Tests that _copy_latest_fault errors are swallowed"""
        with mock.patch('nova.objects.InstanceFault.get_latest_for_instance',
                        side_effect=test.TestingException):
            self.task._copy_latest_fault(self.source_context)
        # The source cell should not have a fault.
        self.assertIsNone(self.task.source_cell_instance.fault)
        # The error should have been logged.
        mock_log.assert_called_once()
        self.assertIn('Failed to copy instance fault from target cell DB',
                      mock_log.call_args[0][0])


class UtilityTestCase(test.NoDBTestCase):
    """Tests utility methods in the cross_cell_migrate module."""

    @mock.patch('nova.objects.HostMapping.get_by_host',
                return_value=objects.HostMapping(
                    cell_mapping=objects.CellMapping(uuid=uuids.cell)))
    @mock.patch('nova.objects.Instance.get_by_uuid')
    def test_get_inst_and_cell_map_from_source(self, mock_get_inst,
                                               mock_get_by_host):
        target_cell_context = nova_context.get_admin_context()
        # Stub out Instance.get_by_uuid to make sure a copy of the context is
        # targeted at the source cell mapping.

        def stub_get_by_uuid(ctxt, *args, **kwargs):
            self.assertIsNot(ctxt, target_cell_context)
            self.assertEqual(uuids.cell, ctxt.cell_uuid)
            return mock.sentinel.instance
        mock_get_inst.side_effect = stub_get_by_uuid
        inst, cell_mapping = (
            cross_cell_migrate.get_inst_and_cell_map_from_source(
                target_cell_context, 'source-host', uuids.instance))
        self.assertIs(inst, mock.sentinel.instance)
        self.assertIs(cell_mapping, mock_get_by_host.return_value.cell_mapping)
        mock_get_by_host.assert_called_once_with(
            target_cell_context, 'source-host')
        mock_get_inst.assert_called_once_with(
            test.MatchType(nova_context.RequestContext), uuids.instance,
            expected_attrs=['flavor', 'info_cache', 'system_metadata'])


class ConfirmResizeTaskTestCase(test.NoDBTestCase):

    def setUp(self):
        super(ConfirmResizeTaskTestCase, self).setUp()
        context = nova_context.get_admin_context()
        compute_rpcapi = mock.Mock()
        self.task = cross_cell_migrate.ConfirmResizeTask(
            context,
            objects.Instance(context, uuid=uuids.instance,
                             host='target-host', vm_state=vm_states.RESIZED,
                             system_metadata={
                                 'old_vm_state': vm_states.ACTIVE}),
            objects.Migration(context, uuid=uuids.migration,
                              dest_compute='target-host',
                              source_compute='source-host',
                              status='confirming'),
            mock.sentinel.legacy_notifier,
            compute_rpcapi)

    @mock.patch('nova.conductor.tasks.cross_cell_migrate.'
                'get_inst_and_cell_map_from_source')
    def test_execute(self, mock_get_instance):
        source_cell_instance = objects.Instance(
            mock.MagicMock(), uuid=uuids.instance)
        source_cell_instance.destroy = mock.Mock()
        mock_get_instance.return_value = (
            source_cell_instance, objects.CellMapping())
        with test.nested(
            mock.patch.object(self.task, '_send_resize_confirm_notification'),
            mock.patch.object(self.task, '_cleanup_source_host'),
            mock.patch.object(self.task, '_finish_confirm_in_target_cell')
        ) as (
            _send_resize_confirm_notification, _cleanup_source_host,
            _finish_confirm_in_target_cell
        ):
            self.task.execute()
        mock_get_instance.assert_called_once_with(
            self.task.context, self.task.migration.source_compute,
            self.task.instance.uuid)
        self.assertEqual(2, _send_resize_confirm_notification.call_count)
        _send_resize_confirm_notification.assert_has_calls([
            mock.call(source_cell_instance, fields.NotificationPhase.START),
            mock.call(self.task.instance, fields.NotificationPhase.END)])
        _cleanup_source_host.assert_called_once_with(source_cell_instance)
        source_cell_instance.destroy.assert_called_once_with(
            hard_delete=True)
        _finish_confirm_in_target_cell.assert_called_once_with()

    @mock.patch('nova.conductor.tasks.cross_cell_migrate.'
                'get_inst_and_cell_map_from_source',
                side_effect=exception.InstanceNotFound(
                    instance_id=uuids.instance))
    @mock.patch('nova.objects.Migration.save')
    @mock.patch('nova.objects.RequestSpec.get_by_instance_uuid')
    @mock.patch('nova.scheduler.utils.set_vm_state_and_notify')
    def test_rollback(self, mock_set_state_notify, mock_get_reqspec,
                      mock_mig_save, mock_get_instance):
        self.assertRaises(exception.InstanceNotFound, self.task.execute)
        mock_get_instance.assert_called_once_with(
            self.task.context, self.task.migration.source_compute,
            self.task.instance.uuid)
        self.assertEqual('error', self.task.migration.status)
        mock_mig_save.assert_called_once_with()
        mock_get_reqspec.assert_called_once_with(
            self.task.context, self.task.instance.uuid)
        mock_set_state_notify.assert_called_once_with(
            self.task.context, self.task.instance.uuid, 'compute_task',
            'migrate_server',
            {'vm_state': vm_states.ERROR, 'task_state': None},
            mock_get_instance.side_effect,
            mock_get_reqspec.return_value)

    @mock.patch('nova.compute.utils.notify_about_instance_usage')
    @mock.patch('nova.compute.utils.notify_about_instance_action')
    def test_send_resize_confirm_notification(self, mock_versioned_notify,
                                              mock_legacy_notify):
        self.flags(host='fake-conductor-host')
        instance = self.task.instance
        self.task._send_resize_confirm_notification(instance, 'fake-phase')
        mock_legacy_notify.assert_called_once_with(
            self.task.legacy_notifier, instance._context, instance,
            'resize.confirm.fake-phase')
        mock_versioned_notify.assert_called_once_with(
            instance._context, instance, 'fake-conductor-host',
            action=fields.NotificationAction.RESIZE_CONFIRM,
            phase='fake-phase')

    @mock.patch('nova.objects.InstanceAction.action_start')
    @mock.patch('nova.objects.Migration.get_by_uuid')
    @mock.patch('nova.objects.InstanceActionEvent')  # stub EventReporter calls
    def test_cleanup_source_host(
            self, mock_action_event, mock_get_mig, mock_action_start):
        instance = objects.Instance(nova_context.get_admin_context(),
                                    uuid=uuids.instance,
                                    flavor=objects.Flavor())
        self.task._cleanup_source_host(instance)
        mock_action_start.assert_called_once_with(
            instance._context, instance.uuid, instance_actions.CONFIRM_RESIZE,
            want_result=False)
        mock_get_mig.assert_called_once_with(
            instance._context, self.task.migration.uuid)
        self.task.compute_rpcapi.confirm_snapshot_based_resize_at_source.\
            assert_called_once_with(instance._context, instance,
                                    mock_get_mig.return_value)
        mock_action_event.event_start.assert_called_once_with(
            self.task.context, uuids.instance,
            'compute_confirm_snapshot_based_resize_at_source',
            want_result=False, host=mock_get_mig.return_value.source_compute)
        mock_action_event.event_finish_with_failure.assert_called_once_with(
            self.task.context, uuids.instance,
            'compute_confirm_snapshot_based_resize_at_source',
            exc_val=None, exc_tb=None, want_result=False)

    @mock.patch('nova.objects.Migration.save')
    @mock.patch('nova.objects.Instance.save')
    @mock.patch('nova.objects.Instance.drop_migration_context')
    def test_finish_confirm_in_target_cell(self, mock_drop_ctx, mock_inst_save,
                                           mock_mig_save):
        with mock.patch.object(
                self.task, '_set_vm_and_task_state') as mock_set_state:
            self.task._finish_confirm_in_target_cell()
        self.assertEqual('confirmed', self.task.migration.status)
        mock_mig_save.assert_called_once_with()
        self.assertNotIn('old_vm_state', self.task.instance.system_metadata)
        self.assertIsNone(self.task.instance.old_flavor)
        self.assertIsNone(self.task.instance.new_flavor)
        mock_set_state.assert_called_once_with()
        mock_drop_ctx.assert_called_once_with()
        mock_inst_save.assert_called_once_with(expected_task_state=[
            None, task_states.DELETING, task_states.SOFT_DELETING])

    def test_set_vm_and_task_state_shutdown(self):
        self.task.instance.power_state = power_state.SHUTDOWN
        self.task._set_vm_and_task_state()
        self.assertEqual(vm_states.STOPPED, self.task.instance.vm_state)
        self.assertIsNone(self.task.instance.task_state)

    def test_set_vm_and_task_state_active(self):
        self.task.instance.power_state = power_state.RUNNING
        self.task._set_vm_and_task_state()
        self.assertEqual(vm_states.ACTIVE, self.task.instance.vm_state)
        self.assertIsNone(self.task.instance.task_state)


class RevertResizeTaskTestCase(test.NoDBTestCase, ObjectComparatorMixin):

    def setUp(self):
        super(RevertResizeTaskTestCase, self).setUp()
        target_cell_context = nova_context.get_admin_context()
        target_cell_context.cell_uuid = uuids.target_cell
        instance = fake_instance.fake_instance_obj(
            target_cell_context, **{
                'vm_state': vm_states.RESIZED,
                'task_state': task_states.RESIZE_REVERTING,
                'expected_attrs': ['system_metadata', 'flavor']
            })
        migration = objects.Migration(
            target_cell_context, uuid=uuids.migration, status='reverting',
            source_compute='source-host', dest_compute='dest-host')
        legacy_notifier = mock.MagicMock()
        compute_rpcapi = mock.MagicMock()
        self.task = cross_cell_migrate.RevertResizeTask(
            target_cell_context, instance, migration, legacy_notifier,
            compute_rpcapi)

    def _generate_source_cell_instance(self):
        source_cell_context = nova_context.get_admin_context()
        source_cell_context.cell_uuid = uuids.source_cell
        source_cell_instance = self.task.instance.obj_clone()
        source_cell_instance._context = source_cell_context
        return source_cell_instance

    @mock.patch('nova.conductor.tasks.cross_cell_migrate.'
                'get_inst_and_cell_map_from_source')
    @mock.patch('nova.objects.InstanceActionEvent')  # Stub EventReport calls.
    def test_execute(self, mock_action_event, mock_get_instance):
        """Happy path test for the execute method."""
        # Setup mocks.
        source_cell_instance = self._generate_source_cell_instance()
        source_cell_context = source_cell_instance._context
        source_cell_mapping = objects.CellMapping(source_cell_context,
                                                  uuid=uuids.source_cell)
        mock_get_instance.return_value = (source_cell_instance,
                                          source_cell_mapping)

        def stub_update_instance_in_source_cell(*args, **kwargs):
            # Ensure _update_instance_mapping is not called before
            # _update_instance_in_source_cell.
            _update_instance_mapping.assert_not_called()
            return mock.sentinel.source_cell_migration

        with test.nested(
            mock.patch.object(self.task, '_send_resize_revert_notification'),
            mock.patch.object(self.task, '_update_instance_in_source_cell',
                              side_effect=stub_update_instance_in_source_cell),
            mock.patch.object(self.task, '_update_instance_mapping'),
            mock.patch.object(self.task.instance, 'destroy'),
            mock.patch.object(source_cell_instance, 'refresh'),
        ) as (
            _send_resize_revert_notification, _update_instance_in_source_cell,
            _update_instance_mapping, mock_inst_destroy, mock_inst_refresh,
        ):
            # Run the code.
            self.task.execute()
        # Should have sent a start and end notification.
        self.assertEqual(2, _send_resize_revert_notification.call_count,
                         _send_resize_revert_notification.calls)
        _send_resize_revert_notification.assert_has_calls([
            mock.call(self.task.instance, fields.NotificationPhase.START),
            mock.call(source_cell_instance, fields.NotificationPhase.END),
        ])
        mock_get_instance.assert_called_once_with(
            self.task.context, self.task.migration.source_compute,
            self.task.instance.uuid)
        _update_instance_in_source_cell.assert_called_once_with(
            source_cell_instance)
        _update_instance_mapping.assert_called_once_with(
            source_cell_instance, source_cell_mapping)
        # _source_cell_instance and _source_cell_migration should have been
        # set for rollbacks
        self.assertIs(self.task._source_cell_instance, source_cell_instance)
        self.assertIs(self.task._source_cell_migration,
                      mock.sentinel.source_cell_migration)
        # Cleanup at dest host.
        self.task.compute_rpcapi.revert_snapshot_based_resize_at_dest.\
            assert_called_once_with(self.task.context, self.task.instance,
                                    self.task.migration)
        # EventReporter should have been used.
        event_name = 'compute_revert_snapshot_based_resize_at_dest'
        mock_action_event.event_start.assert_called_once_with(
            source_cell_context, source_cell_instance.uuid, event_name,
            want_result=False, host=self.task.migration.dest_compute)
        mock_action_event.event_finish_with_failure.assert_called_once_with(
            source_cell_context, source_cell_instance.uuid, event_name,
            exc_val=None, exc_tb=None, want_result=False)
        mock_action_event.event_finish.assert_called_once_with(
            source_cell_context, source_cell_instance.uuid,
            'conductor_revert_snapshot_based_resize', want_result=False)
        # Destroy the instance in the target cell.
        mock_inst_destroy.assert_called_once_with(hard_delete=True)
        # Cleanup at source host.
        self.task.compute_rpcapi.\
            finish_revert_snapshot_based_resize_at_source.\
            assert_called_once_with(
                source_cell_context, source_cell_instance,
                mock.sentinel.source_cell_migration)
        # Refresh the source cell instance so we have the latest data.
        mock_inst_refresh.assert_called_once_with()

    @mock.patch('nova.conductor.tasks.cross_cell_migrate.RevertResizeTask.'
                '_execute')
    @mock.patch('nova.objects.RequestSpec.get_by_instance_uuid')
    @mock.patch('nova.scheduler.utils.set_vm_state_and_notify')
    def test_rollback_target_cell(
            self, mock_set_state_notify, mock_get_reqspec, mock_execute):
        """Tests the case that we did not update the instance mapping
        so we set the target cell migration to error status.
        """
        error = test.TestingException('zoinks!')
        mock_execute.side_effect = error
        with mock.patch.object(self.task.migration, 'save') as mock_save:
            self.assertRaises(test.TestingException, self.task.execute)
        self.assertEqual('error', self.task.migration.status)
        mock_save.assert_called_once_with()
        mock_get_reqspec.assert_called_once_with(
            self.task.context, self.task.instance.uuid)
        mock_set_state_notify.assert_called_once_with(
            self.task.instance._context, self.task.instance.uuid,
            'compute_task', 'migrate_server',
            {'vm_state': vm_states.ERROR, 'task_state': None}, error,
            mock_get_reqspec.return_value)
        self.assertIn('The instance is mapped to the target cell',
                      self.stdlog.logger.output)

    @mock.patch('nova.conductor.tasks.cross_cell_migrate.RevertResizeTask.'
                '_execute')
    @mock.patch('nova.objects.RequestSpec.get_by_instance_uuid')
    @mock.patch('nova.scheduler.utils.set_vm_state_and_notify')
    def test_rollback_source_cell(
            self, mock_set_state_notify, mock_get_reqspec, mock_execute):
        """Tests the case that we did update the instance mapping
        so we set the source cell migration to error status.
        """
        source_cell_instance = self._generate_source_cell_instance()
        source_cell_context = source_cell_instance._context
        self.task._source_cell_instance = source_cell_instance
        self.task._source_cell_migration = objects.Migration(
            source_cell_context, status='reverting', dest_compute='dest-host')
        error = test.TestingException('jinkies!')
        mock_execute.side_effect = error
        with mock.patch.object(self.task._source_cell_migration,
                               'save') as mock_save:
            self.assertRaises(test.TestingException, self.task.execute)
        self.assertEqual('error', self.task._source_cell_migration.status)
        mock_save.assert_called_once_with()
        mock_get_reqspec.assert_called_once_with(
            self.task.context, self.task.instance.uuid)
        mock_set_state_notify.assert_called_once_with(
            source_cell_context, source_cell_instance.uuid,
            'compute_task', 'migrate_server',
            {'vm_state': vm_states.ERROR, 'task_state': None}, error,
            mock_get_reqspec.return_value)
        self.assertIn('The instance is mapped to the source cell',
                      self.stdlog.logger.output)

    @mock.patch('nova.compute.utils.notify_about_instance_usage')
    @mock.patch('nova.compute.utils.notify_about_instance_action')
    def test_send_resize_revert_notification(self, mock_notify_action,
                                             mock_notify_usage):
        self.flags(host='fake-conductor-host')
        instance = self.task.instance
        self.task._send_resize_revert_notification(instance, 'foo')
        # Assert the legacy notification was sent.
        mock_notify_usage.assert_called_once_with(
            self.task.legacy_notifier, instance._context, instance,
            'resize.revert.foo')
        # Assert the versioned notification was sent.
        mock_notify_action.assert_called_once_with(
            instance._context, instance, 'fake-conductor-host',
            action=fields.NotificationAction.RESIZE_REVERT, phase='foo')

    def test_update_instance_in_source_cell(self):
        # Setup mocks.
        source_cell_instance = self._generate_source_cell_instance()
        source_cell_instance.task_state = None
        self.task.instance.system_metadata = {'old_vm_state': vm_states.ACTIVE}
        with test.nested(
            mock.patch.object(source_cell_instance, 'save'),
            mock.patch.object(self.task, '_update_bdms_in_source_cell'),
            mock.patch.object(self.task,
                              '_update_instance_actions_in_source_cell'),
            mock.patch.object(self.task, '_update_migration_in_source_cell')
        ) as (
            mock_inst_save, _update_bdms_in_source_cell,
            _update_instance_actions_in_source_cell,
            _update_migration_in_source_cell
        ):
            # Run the code.
            source_cell_migration = self.task._update_instance_in_source_cell(
                source_cell_instance)
        # The returned object should be the updated migration object from the
        # source cell database.
        self.assertIs(source_cell_migration,
                      _update_migration_in_source_cell.return_value)
        # Fields on the source cell instance should have been updated.
        self.assertEqual(vm_states.ACTIVE,
                         source_cell_instance.system_metadata['old_vm_state'])
        self.assertEqual(task_states.RESIZE_REVERTING,
                         source_cell_instance.task_state)
        mock_inst_save.assert_called_once_with()
        _update_bdms_in_source_cell.assert_called_once_with(
            source_cell_instance._context)
        _update_instance_actions_in_source_cell.assert_called_once_with(
            source_cell_instance._context)
        _update_migration_in_source_cell.assert_called_once_with(
            source_cell_instance._context)

    @mock.patch('nova.objects.BlockDeviceMappingList.get_by_instance_uuid')
    @mock.patch('nova.conductor.tasks.cross_cell_migrate.'
                'clone_creatable_object')
    def test_update_bdms_in_source_cell(self, mock_clone, mock_get_bdms):
        """Test updating BDMs from the target cell to the source cell."""
        source_cell_context = nova_context.get_admin_context()
        source_cell_context.cell_uuid = uuids.source_cell
        # Setup fake bdms.
        bdm1 = objects.BlockDeviceMapping(
            source_cell_context, uuid=uuids.bdm1, volume_id='vol1',
            attachment_id=uuids.attach1)
        bdm2 = objects.BlockDeviceMapping(
            source_cell_context, uuid=uuids.bdm2, volume_id='vol2',
            attachment_id=uuids.attach2)
        source_bdms = objects.BlockDeviceMappingList(objects=[bdm1, bdm2])
        # With the target BDMs bdm1 from the source is gone and bdm3 is new
        # to simulate bdm1 being detached and bdm3 being attached while the
        # instance was in VERIFY_RESIZE status.
        bdm3 = objects.BlockDeviceMapping(
            self.task.context, uuid=uuids.bdm3, volume_id='vol3',
            attachment_id=uuids.attach1)
        target_bdms = objects.BlockDeviceMappingList(objects=[bdm2, bdm3])

        def stub_get_bdms(ctxt, *args, **kwargs):
            if ctxt.cell_uuid == uuids.source_cell:
                return source_bdms
            return target_bdms
        mock_get_bdms.side_effect = stub_get_bdms

        def stub_mock_clone(ctxt, obj, *args, **kwargs):
            # We want to make assertions on our mocks so do not create a copy.
            return obj
        mock_clone.side_effect = stub_mock_clone

        with test.nested(
            mock.patch.object(self.task.volume_api, 'attachment_create',
                              return_value={'id': uuids.attachment_id}),
            mock.patch.object(self.task.volume_api, 'attachment_delete'),
            mock.patch.object(bdm3, 'create'),
            mock.patch.object(bdm1, 'destroy')
        ) as (
            mock_attachment_create, mock_attachment_delete,
            mock_bdm_create, mock_bdm_destroy
        ):
            self.task._update_bdms_in_source_cell(source_cell_context)
        # Should have gotten BDMs from the source and target cell (order does
        # not matter).
        self.assertEqual(2, mock_get_bdms.call_count, mock_get_bdms.calls)
        mock_get_bdms.assert_has_calls([
            mock.call(source_cell_context, self.task.instance.uuid),
            mock.call(self.task.context, self.task.instance.uuid)],
            any_order=True)
        # Since bdm3 was new in the target cell an attachment should have been
        # created for it in the source cell.
        mock_attachment_create.assert_called_once_with(
            source_cell_context, bdm3.volume_id, self.task.instance.uuid)
        self.assertEqual(uuids.attachment_id, bdm3.attachment_id)
        # And bdm3 should have been created in the source cell.
        mock_bdm_create.assert_called_once_with()
        # Since bdm1 was not in the target cell it should be destroyed in the
        # source cell since we can assume it was detached from the target host
        # in the target cell while the instance was in VERIFY_RESIZE status.
        mock_attachment_delete.assert_called_once_with(
            bdm1._context, bdm1.attachment_id)
        mock_bdm_destroy.assert_called_once_with()

    @mock.patch('nova.objects.BlockDeviceMapping.destroy')
    def test_delete_orphan_source_cell_bdms_attach_delete_fails(self, destroy):
        """Tests attachment_delete failing but not being fatal."""
        source_cell_context = nova_context.get_admin_context()
        bdm1 = objects.BlockDeviceMapping(
            source_cell_context, volume_id='vol1', attachment_id=uuids.attach1)
        bdm2 = objects.BlockDeviceMapping(
            source_cell_context, volume_id='vol2', attachment_id=uuids.attach2)
        source_cell_bdms = objects.BlockDeviceMappingList(objects=[bdm1, bdm2])
        with mock.patch.object(self.task.volume_api,
                               'attachment_delete') as attachment_delete:
            # First call to attachment_delete fails, second is OK.
            attachment_delete.side_effect = [
                test.TestingException('cinder is down'), None]
            self.task._delete_orphan_source_cell_bdms(source_cell_bdms)
        attachment_delete.assert_has_calls([
            mock.call(bdm1._context, bdm1.attachment_id),
            mock.call(bdm2._context, bdm2.attachment_id)])
        self.assertEqual(2, destroy.call_count, destroy.mock_calls)
        self.assertIn('cinder is down', self.stdlog.logger.output)

    @mock.patch('nova.objects.InstanceAction.get_by_request_id')
    @mock.patch('nova.objects.InstanceActionEventList.get_by_action')
    @mock.patch('nova.objects.InstanceAction.create')
    @mock.patch('nova.objects.InstanceActionEvent.create')
    def test_update_instance_actions_in_source_cell(
            self, mock_event_create, mock_action_create, mock_get_events,
            mock_get_action):
        """Tests copying instance actions from the target to source cell."""
        source_cell_context = nova_context.get_admin_context()
        source_cell_context.cell_uuid = uuids.source_cell
        # Setup a fake action and fake event.
        action = objects.InstanceAction(
            id=1, action=instance_actions.REVERT_RESIZE,
            instance_uuid=self.task.instance.uuid,
            request_id=self.task.context.request_id)
        mock_get_action.return_value = action
        event = objects.InstanceActionEvent(
            id=2, action_id=action.id,
            event='conductor_revert_snapshot_based_resize')
        mock_get_events.return_value = objects.InstanceActionEventList(
            objects=[event])
        # Run the code.
        self.task._update_instance_actions_in_source_cell(source_cell_context)
        # Should have created a clone of the action and event.
        mock_get_action.assert_called_once_with(
            self.task.context, self.task.instance.uuid,
            self.task.context.request_id)
        mock_get_events.assert_called_once_with(self.task.context, action.id)
        mock_action_create.assert_called_once_with()
        mock_event_create.assert_called_once_with(
            action.instance_uuid, action.request_id)

    def test_update_source_obj_from_target_cell(self):
        # Create a fake source object to be updated.
        t1 = timeutils.utcnow()
        source_obj = objects.Migration(id=1, created_at=t1, updated_at=t1,
                                       uuid=uuids.migration,
                                       status='post-migrating')
        t2 = timeutils.utcnow()
        target_obj = objects.Migration(id=2, created_at=t2, updated_at=t2,
                                       uuid=uuids.migration,
                                       status='reverting',
                                       # Add a field that is not in source_obj.
                                       migration_type='resize')
        # Run the copy code.
        self.task._update_source_obj_from_target_cell(source_obj, target_obj)
        # First make sure that id, created_at and updated_at are not changed.
        ignored_keys = ['id', 'created_at', 'updated_at']
        for field in ignored_keys:
            self.assertNotEqual(getattr(source_obj, field),
                                getattr(target_obj, field))
        # Now make sure the rest of the fields are the same.
        self._compare_objs(source_obj, target_obj, ignored_keys=ignored_keys)

    def test_update_source_obj_from_target_cell_nested_object(self):
        """Tests that calling _update_source_obj_from_target_cell with an
        object that has nested object fields will raise ObjectActionError.
        """
        source = objects.Instance(flavor=objects.Flavor(flavorid='a'))
        target = objects.Instance(flavor=objects.Flavor(flavorid='b'))
        ex = self.assertRaises(exception.ObjectActionError,
                               self.task._update_source_obj_from_target_cell,
                               source, target)
        self.assertIn('nested objects are not supported', str(ex))

    @mock.patch('nova.objects.Migration.get_by_uuid')
    def test_update_migration_in_source_cell(self, mock_get_migration):
        """Tests updating the migration record in the source cell from the
        target cell.
        """
        source_cell_context = nova_context.get_admin_context()
        with mock.patch.object(
                self.task,
                '_update_source_obj_from_target_cell') as mock_update_obj:
            source_cell_migration = \
                self.task._update_migration_in_source_cell(source_cell_context)
        mock_get_migration.assert_called_once_with(source_cell_context,
                                                   self.task.migration.uuid)
        mock_update_obj.assert_called_once_with(source_cell_migration,
                                                self.task.migration)
        self.assertIs(source_cell_migration, mock_get_migration.return_value)
        source_cell_migration.save.assert_called_once_with()

    @mock.patch('nova.objects.InstanceMapping.get_by_instance_uuid')
    def test_update_instance_mapping(self, get_inst_map):
        """Tests updating the instance mapping from the target to source cell.
        """
        source_cell_instance = self._generate_source_cell_instance()
        source_cell_context = source_cell_instance._context
        source_cell_mapping = objects.CellMapping(source_cell_context,
                                                  uuid=uuids.source_cell)
        inst_map = objects.InstanceMapping(
            cell_mapping=objects.CellMapping(uuids.target_cell))
        get_inst_map.return_value = inst_map
        with test.nested(
            mock.patch.object(source_cell_instance, 'save'),
            mock.patch.object(self.task.instance, 'save'),
            mock.patch.object(inst_map, 'save')
        ) as (
            source_inst_save, target_inst_save, inst_map_save
        ):
            self.task._update_instance_mapping(source_cell_instance,
                                               source_cell_mapping)
        get_inst_map.assert_called_once_with(self.task.context,
                                             self.task.instance.uuid)
        # The source cell instance should not be hidden.
        self.assertFalse(source_cell_instance.hidden)
        source_inst_save.assert_called_once_with()
        # The instance mapping should point at the source cell.
        self.assertIs(source_cell_mapping, inst_map.cell_mapping)
        inst_map_save.assert_called_once_with()
        # The target cell instance should be hidden.
        self.assertTrue(self.task.instance.hidden)
        target_inst_save.assert_called_once_with()



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import fixture as config_fixture
from oslo_policy import opts as policy_opts

from nova.conf import devices
from nova.conf import neutron
from nova.conf import paths
from nova import config


class ConfFixture(config_fixture.Config):
    """Fixture to manage global conf settings."""

    def setUp(self):
        super(ConfFixture, self).setUp()

        # default group
        self.conf.set_default('compute_driver', 'fake.SmallFakeDriver')
        self.conf.set_default('host', 'fake-mini')
        self.conf.set_default('periodic_enable', False)

        # api_database group
        self.conf.set_default('connection', "sqlite://", group='api_database')
        self.conf.set_default('sqlite_synchronous', False,
                              group='api_database')

        # database group
        self.conf.set_default('connection', "sqlite://", group='database')
        self.conf.set_default('sqlite_synchronous', False, group='database')

        # key_manager group
        self.conf.set_default('backend',
                              'nova.keymgr.conf_key_mgr.ConfKeyManager',
                              group='key_manager')

        # wsgi group
        self.conf.set_default('api_paste_config',
                              paths.state_path_def('etc/nova/api-paste.ini'),
                              group='wsgi')
        # The functional tests run wsgi API services using fixtures and
        # eventlet and we want one connection per request so things don't
        # leak between requests from separate services in concurrently running
        # tests.
        self.conf.set_default('keep_alive', False, group="wsgi")

        # many tests synchronizes on the reception of versioned notifications
        self.conf.set_default(
            'notification_format', "both", group="notifications")

        # oslo.limit requires endpoint_id since 2.3.0
        self.conf.set_default('endpoint_id', 'ENDPOINT_ID', group='oslo_limit')

        config.parse_args([], default_config_files=[], configure_db=False,
                          init_rpc=False)
        policy_opts.set_defaults(self.conf)
        neutron.register_dynamic_opts(self.conf)
        devices.register_dynamic_opts(self.conf)



#    Copyright 2013 Rackspace Hosting.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from nova import context
from nova.db.main import api as db_api
from nova import exception
from nova.objects import quotas as quotas_obj
from nova import quota
from nova import test
from nova.tests.unit import fake_instance
from nova.tests.unit.objects import test_objects


QUOTAS = quota.QUOTAS


class TestQuotasModule(test.NoDBTestCase):
    def setUp(self):
        super(TestQuotasModule, self).setUp()
        self.context = context.RequestContext('fake_user1', 'fake_proj1')
        self.instance = fake_instance.fake_db_instance(
                project_id='fake_proj2', user_id='fake_user2')

    def test_ids_from_instance_non_admin(self):
        project_id, user_id = quotas_obj.ids_from_instance(
                self.context, self.instance)
        self.assertEqual('fake_user2', user_id)
        self.assertEqual('fake_proj1', project_id)

    def test_ids_from_instance_admin(self):
        project_id, user_id = quotas_obj.ids_from_instance(
                self.context.elevated(), self.instance)
        self.assertEqual('fake_user2', user_id)
        self.assertEqual('fake_proj2', project_id)


class _TestQuotasObject(object):
    def setUp(self):
        super(_TestQuotasObject, self).setUp()
        self.context = context.RequestContext('fake_user1', 'fake_proj1')
        self.instance = fake_instance.fake_db_instance(
                project_id='fake_proj2', user_id='fake_user2')

    @mock.patch(
        'nova.db.main.api.quota_get', side_effect=exception.QuotaNotFound)
    @mock.patch('nova.objects.Quotas._create_limit_in_db')
    def test_create_limit(self, mock_create, mock_get):
        quotas_obj.Quotas.create_limit(self.context, 'fake-project',
                                       'foo', 10, user_id='user')
        mock_create.assert_called_once_with(self.context, 'fake-project',
                                            'foo', 10, user_id='user')

    @mock.patch('nova.db.main.api.quota_get')
    @mock.patch('nova.objects.Quotas._create_limit_in_db')
    def test_create_limit_exists_in_main(self, mock_create, mock_get):
        self.assertRaises(exception.QuotaExists,
                          quotas_obj.Quotas.create_limit, self.context,
                          'fake-project', 'foo', 10, user_id='user')

    @mock.patch('nova.objects.Quotas._update_limit_in_db')
    def test_update_limit(self, mock_update):
        quotas_obj.Quotas.update_limit(self.context, 'fake-project',
                                       'foo', 10, user_id='user')
        mock_update.assert_called_once_with(self.context, 'fake-project',
                                            'foo', 10, user_id='user')

    @mock.patch.object(QUOTAS, 'count_as_dict')
    def test_count(self, mock_count):
        # key_pairs can't actually be counted across a project, this is just
        # for testing.
        mock_count.return_value = {'project': {'key_pairs': 5},
                                   'user': {'key_pairs': 4}}
        count = quotas_obj.Quotas.count(self.context, 'key_pairs', 'a-user')
        self.assertEqual(4, count)

        # key_pairs can't actually be counted across a project, this is just
        # for testing.
        mock_count.return_value = {'project': {'key_pairs': 5}}
        count = quotas_obj.Quotas.count(self.context, 'key_pairs', 'a-user')
        self.assertEqual(5, count)

        mock_count.return_value = {'user': {'key_pairs': 3}}
        count = quotas_obj.Quotas.count(self.context, 'key_pairs', 'a-user')
        self.assertEqual(3, count)

    @mock.patch('nova.objects.Quotas.count_as_dict')
    def test_check_deltas(self, mock_count):
        self.flags(key_pairs=3, group='quota')
        self.flags(server_group_members=3, group='quota')

        def fake_count(context, resource):
            if resource in ('key_pairs', 'server_group_members'):
                return {'project': {'key_pairs': 2, 'server_group_members': 2},
                        'user': {'key_pairs': 1, 'server_group_members': 2}}
            else:
                return {'user': {resource: 2}}

        mock_count.side_effect = fake_count
        deltas = {'key_pairs': 1,
                  'server_group_members': 1,
                  'security_group_rules': 1}
        project_id = 'fake-other-project'
        user_id = 'fake-other-user'
        quotas_obj.Quotas.check_deltas(self.context, deltas,
                                       check_project_id=project_id,
                                       check_user_id=user_id)
        # Should be called twice: once for key_pairs/server_group_members,
        # once for security_group_rules.
        self.assertEqual(2, mock_count.call_count)
        call1 = mock.call(self.context, 'key_pairs')
        call2 = mock.call(self.context, 'server_group_members')
        call3 = mock.call(self.context, 'security_group_rules')
        self.assertTrue(call1 in mock_count.mock_calls or
                        call2 in mock_count.mock_calls)
        self.assertIn(call3, mock_count.mock_calls)

    @mock.patch('nova.objects.Quotas.count_as_dict')
    def test_check_deltas_zero(self, mock_count):
        # This will test that we will raise OverQuota if given a zero delta if
        # an object creation has put us over the allowed quota.
        # This is for the scenario where we recheck quota and delete an object
        # if we have gone over quota during a race.
        self.flags(key_pairs=3, group='quota')
        self.flags(server_group_members=3, group='quota')

        def fake_count(context, resource):
            return {'user': {resource: 4}}

        mock_count.side_effect = fake_count
        deltas = {'key_pairs': 0, 'server_group_members': 0}
        project_id = 'fake-other-project'
        user_id = 'fake-other-user'
        self.assertRaises(exception.OverQuota, quotas_obj.Quotas.check_deltas,
                          self.context, deltas,
                          check_project_id=project_id,
                          check_user_id=user_id)
        # Should be called twice, once for key_pairs, once for
        # server_group_members
        self.assertEqual(2, mock_count.call_count)
        call1 = mock.call(self.context, 'key_pairs')
        call2 = mock.call(self.context, 'server_group_members')
        mock_count.assert_has_calls([call1, call2], any_order=True)

    @mock.patch('nova.objects.Quotas.count_as_dict')
    def test_check_deltas_negative(self, mock_count):
        """Test check_deltas with a negative delta.

        Negative deltas probably won't be used going forward for countable
        resources because there are no usage records to decrement and there
        won't be quota operations done when deleting resources. When resources
        are deleted, they will no longer be reflected in the count.
        """
        self.flags(key_pairs=3, group='quota')
        mock_count.return_value = {'user': {'key_pairs': 4}}
        deltas = {'key_pairs': -1}
        # Should pass because the delta makes 3 key_pairs
        quotas_obj.Quotas.check_deltas(self.context, deltas, 'a-user',
                                       something='something')
        # args for the count function should get passed along
        mock_count.assert_called_once_with(self.context, 'key_pairs', 'a-user',
                                           something='something')

    @mock.patch('nova.objects.Quotas.count_as_dict')
    @mock.patch('nova.objects.Quotas.limit_check_project_and_user')
    def test_check_deltas_limit_check_scoping(self, mock_check, mock_count):
        # check_project_id and check_user_id kwargs should get passed along to
        # limit_check_project_and_user()
        mock_count.return_value = {'project': {'foo': 5}, 'user': {'foo': 1}}
        deltas = {'foo': 1}

        quotas_obj.Quotas.check_deltas(self.context, deltas, 'a-project')
        mock_check.assert_called_once_with(self.context,
                                           project_values={'foo': 6},
                                           user_values={'foo': 2})

        mock_check.reset_mock()
        quotas_obj.Quotas.check_deltas(self.context, deltas, 'a-project',
                                       check_project_id='a-project')
        mock_check.assert_called_once_with(self.context,
                                           project_values={'foo': 6},
                                           user_values={'foo': 2},
                                           project_id='a-project')

        mock_check.reset_mock()
        quotas_obj.Quotas.check_deltas(self.context, deltas, 'a-project',
                                       check_user_id='a-user')
        mock_check.assert_called_once_with(self.context,
                                           project_values={'foo': 6},
                                           user_values={'foo': 2},
                                           user_id='a-user')

    @mock.patch('nova.objects.Quotas._update_limit_in_db',
                side_effect=exception.QuotaNotFound)
    @mock.patch('nova.db.main.api.quota_update')
    def test_update_limit_main(self, mock_update_main, mock_update):
        quotas_obj.Quotas.update_limit(self.context, 'fake-project',
                                       'foo', 10, user_id='user')
        mock_update.assert_called_once_with(self.context, 'fake-project',
                                            'foo', 10, user_id='user')
        mock_update_main.assert_called_once_with(self.context, 'fake-project',
                                                 'foo', 10,
                                                 user_id='user')

    @mock.patch('nova.objects.Quotas._get_from_db')
    def test_get(self, mock_get):
        qclass = quotas_obj.Quotas.get(self.context, 'fake-project', 'foo',
                                       user_id='user')
        mock_get.assert_called_once_with(self.context, 'fake-project', 'foo',
                                         user_id='user')
        self.assertEqual(mock_get.return_value, qclass)

    @mock.patch('nova.objects.Quotas._get_from_db',
                side_effect=exception.QuotaNotFound)
    @mock.patch('nova.db.main.api.quota_get')
    def test_get_main(self, mock_get_main, mock_get):
        quotas_obj.Quotas.get(self.context, 'fake-project', 'foo',
                              user_id='user')
        mock_get.assert_called_once_with(self.context, 'fake-project', 'foo',
                                         user_id='user')
        mock_get_main.assert_called_once_with(self.context, 'fake-project',
                                              'foo', user_id='user')

    @mock.patch('nova.objects.Quotas._get_all_from_db')
    @mock.patch('nova.db.main.api.quota_get_all')
    def test_get_all(self, mock_get_all_main, mock_get_all):
        mock_get_all.return_value = ['api1']
        mock_get_all_main.return_value = ['main1', 'main2']
        quotas = quotas_obj.Quotas.get_all(self.context, 'fake-project')
        mock_get_all.assert_called_once_with(self.context, 'fake-project')
        mock_get_all_main.assert_called_once_with(self.context, 'fake-project')
        self.assertEqual(['api1', 'main1', 'main2'], quotas)

    @mock.patch('nova.objects.Quotas._get_all_from_db_by_project')
    @mock.patch('nova.db.main.api.quota_get_all_by_project')
    def test_get_all_by_project(self, mock_get_all_main, mock_get_all):
        mock_get_all.return_value = {'project_id': 'fake-project',
                                     'fixed_ips': 20, 'floating_ips': 5}
        mock_get_all_main.return_value = {'project_id': 'fake-project',
                                          'fixed_ips': 10}
        quotas_dict = quotas_obj.Quotas.get_all_by_project(self.context,
                                                           'fake-project')
        mock_get_all.assert_called_once_with(self.context, 'fake-project')
        mock_get_all_main.assert_called_once_with(self.context, 'fake-project')
        expected = {'project_id': 'fake-project', 'fixed_ips': 20,
                    'floating_ips': 5}
        self.assertEqual(expected, quotas_dict)

    @mock.patch('nova.objects.Quotas._get_all_from_db_by_project_and_user')
    @mock.patch('nova.db.main.api.quota_get_all_by_project_and_user')
    def test_get_all_by_project_and_user(self, mock_get_all_main,
                                         mock_get_all):
        mock_get_all.return_value = {'project_id': 'fake-project',
                                     'user_id': 'user', 'instances': 5,
                                     'cores': 10}
        mock_get_all_main.return_value = {'project_id': 'fake-project',
                                          'user_id': 'user', 'instances': 10,
                                          'ram': 8192}
        quotas_dict = quotas_obj.Quotas.get_all_by_project_and_user(
                self.context, 'fake-project', 'user')
        mock_get_all.assert_called_once_with(self.context, 'fake-project',
                                             'user')
        mock_get_all_main.assert_called_once_with(self.context, 'fake-project',
                                                  'user')
        expected = {'project_id': 'fake-project', 'user_id': 'user',
                    'instances': 5, 'cores': 10, 'ram': 8192}
        self.assertEqual(expected, quotas_dict)

    @mock.patch('nova.objects.Quotas._destroy_all_in_db_by_project')
    def test_destroy_all_by_project(self, mock_destroy_all):
        quotas_obj.Quotas.destroy_all_by_project(self.context, 'fake-project')
        mock_destroy_all.assert_called_once_with(self.context, 'fake-project')

    @mock.patch('nova.objects.Quotas._destroy_all_in_db_by_project',
                side_effect=exception.ProjectQuotaNotFound(
                    project_id='fake-project'))
    @mock.patch('nova.db.main.api.quota_destroy_all_by_project')
    def test_destroy_all_by_project_main(self, mock_destroy_all_main,
                                         mock_destroy_all):
        quotas_obj.Quotas.destroy_all_by_project(self.context, 'fake-project')
        mock_destroy_all.assert_called_once_with(self.context, 'fake-project')
        mock_destroy_all_main.assert_called_once_with(self.context,
                                                      'fake-project')

    @mock.patch('nova.objects.Quotas._destroy_all_in_db_by_project_and_user')
    def test_destroy_all_by_project_and_user(self, mock_destroy_all):
        quotas_obj.Quotas.destroy_all_by_project_and_user(self.context,
                                                          'fake-project',
                                                          'user')
        mock_destroy_all.assert_called_once_with(self.context, 'fake-project',
                                                 'user')

    @mock.patch('nova.objects.Quotas._destroy_all_in_db_by_project_and_user',
                side_effect=exception.ProjectUserQuotaNotFound(
                    user_id='user', project_id='fake-project'))
    @mock.patch('nova.db.main.api.quota_destroy_all_by_project_and_user')
    def test_destroy_all_by_project_and_user_main(self, mock_destroy_all_main,
                                                  mock_destroy_all):
        quotas_obj.Quotas.destroy_all_by_project_and_user(self.context,
                                                          'fake-project',
                                                          'user')
        mock_destroy_all.assert_called_once_with(self.context, 'fake-project',
                                                 'user')
        mock_destroy_all_main.assert_called_once_with(self.context,
                                                      'fake-project', 'user')

    @mock.patch('nova.objects.Quotas._get_class_from_db')
    def test_get_class(self, mock_get):
        qclass = quotas_obj.Quotas.get_class(self.context, 'class', 'resource')
        mock_get.assert_called_once_with(self.context, 'class', 'resource')
        self.assertEqual(mock_get.return_value, qclass)

    @mock.patch('nova.objects.Quotas._get_class_from_db',
                side_effect=exception.QuotaClassNotFound(class_name='class'))
    @mock.patch('nova.db.main.api.quota_class_get')
    def test_get_class_main(self, mock_get_main, mock_get):
        qclass = quotas_obj.Quotas.get_class(self.context, 'class', 'resource')
        mock_get.assert_called_once_with(self.context, 'class', 'resource')
        mock_get_main.assert_called_once_with(self.context, 'class',
                                              'resource')
        self.assertEqual(mock_get_main.return_value, qclass)

    @mock.patch('nova.objects.Quotas._get_all_class_from_db_by_name')
    def test_get_default_class(self, mock_get_all):
        qclass = quotas_obj.Quotas.get_default_class(self.context)
        mock_get_all.assert_called_once_with(self.context,
                                             db_api._DEFAULT_QUOTA_NAME)
        self.assertEqual(mock_get_all.return_value, qclass)

    @mock.patch('nova.objects.Quotas._get_all_class_from_db_by_name',
                side_effect=exception.QuotaClassNotFound(class_name='class'))
    @mock.patch('nova.db.main.api.quota_class_get_default')
    def test_get_default_class_main(self, mock_get_default_main, mock_get_all):
        qclass = quotas_obj.Quotas.get_default_class(self.context)
        mock_get_all.assert_called_once_with(self.context,
                                             db_api._DEFAULT_QUOTA_NAME)
        mock_get_default_main.assert_called_once_with(self.context)
        self.assertEqual(mock_get_default_main.return_value, qclass)

    @mock.patch('nova.objects.Quotas._get_all_class_from_db_by_name')
    @mock.patch('nova.db.main.api.quota_class_get_all_by_name')
    def test_get_class_by_name(self, mock_get_all_main, mock_get_all):
        mock_get_all.return_value = {'class_name': 'foo', 'cores': 10,
                                     'instances': 5}
        mock_get_all_main.return_value = {'class_name': 'foo', 'cores': 20,
                                          'fixed_ips': 10}
        quotas_dict = quotas_obj.Quotas.get_all_class_by_name(self.context,
                                                              'foo')
        mock_get_all.assert_called_once_with(self.context, 'foo')
        mock_get_all_main.assert_called_once_with(self.context, 'foo')
        expected = {'class_name': 'foo', 'cores': 10, 'instances': 5,
                    'fixed_ips': 10}
        self.assertEqual(expected, quotas_dict)

    @mock.patch('nova.db.main.api.quota_class_get',
                side_effect=exception.QuotaClassNotFound(class_name='class'))
    @mock.patch('nova.objects.Quotas._create_class_in_db')
    def test_create_class(self, mock_create, mock_get):
        quotas_obj.Quotas.create_class(self.context, 'class', 'resource',
                                       'limit')
        mock_create.assert_called_once_with(self.context, 'class', 'resource',
                                            'limit')

    @mock.patch('nova.db.main.api.quota_class_get')
    @mock.patch('nova.objects.Quotas._create_class_in_db')
    def test_create_class_exists_in_main(self, mock_create, mock_get):
        self.assertRaises(exception.QuotaClassExists,
                          quotas_obj.Quotas.create_class, self.context,
                          'class', 'resource', 'limit')

    @mock.patch('nova.objects.Quotas._update_class_in_db')
    def test_update_class(self, mock_update):
        quotas_obj.Quotas.update_class(self.context, 'class', 'resource',
                                       'limit')
        mock_update.assert_called_once_with(self.context, 'class', 'resource',
                                            'limit')

    @mock.patch('nova.objects.Quotas._update_class_in_db',
                side_effect=exception.QuotaClassNotFound(class_name='class'))
    @mock.patch('nova.db.main.api.quota_class_update')
    def test_update_class_main(self, mock_update_main, mock_update):
        quotas_obj.Quotas.update_class(self.context, 'class', 'resource',
                                       'limit')
        mock_update.assert_called_once_with(self.context, 'class', 'resource',
                                            'limit')
        mock_update_main.assert_called_once_with(self.context, 'class',
                                                 'resource', 'limit')


class TestQuotasObject(_TestQuotasObject, test_objects._LocalTest):
    pass


class TestRemoteQuotasObject(_TestQuotasObject, test_objects._RemoteTest):
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import typing as ty

from oslo_log import log as logging

from nova import exception
from nova import filesystem
import nova.privsep
from nova.virt import hardware

LOG = logging.getLogger(__name__)

AVAILABLE_PATH = '/sys/devices/system/cpu/present'

CPU_PATH_TEMPLATE = '/sys/devices/system/cpu/cpu%(core)s'


def get_available_cores() -> ty.Set[int]:
    cores = filesystem.read_sys(AVAILABLE_PATH)
    return hardware.parse_cpu_spec(cores) if cores else set()


def exists(core: int) -> bool:
    return core in get_available_cores()


def gen_cpu_path(core: int) -> str:
    if not exists(core):
        LOG.warning('Unable to access CPU: %s', core)
        raise ValueError('CPU: %(core)s does not exist', core)
    return CPU_PATH_TEMPLATE % {'core': core}


def get_online(core: int) -> bool:
    try:
        online = filesystem.read_sys(
            os.path.join(gen_cpu_path(core), 'online')).strip()
    except exception.FileNotFound:
        # The online file may not exist if we haven't written it yet.
        # By default, this means that the CPU is online.
        online = '1'
    return online == '1'


@nova.privsep.sys_admin_pctxt.entrypoint
def set_online(core: int) -> bool:
    filesystem.write_sys(os.path.join(gen_cpu_path(core), 'online'), data='1')
    return get_online(core)


@nova.privsep.sys_admin_pctxt.entrypoint
def set_offline(core: int) -> bool:
    filesystem.write_sys(os.path.join(gen_cpu_path(core), 'online'), data='0')
    return not get_online(core)


def get_governor(core: int) -> str:
    return filesystem.read_sys(
        os.path.join(gen_cpu_path(core), 'cpufreq/scaling_governor')).strip()


@nova.privsep.sys_admin_pctxt.entrypoint
def set_governor(core: int, governor: str) -> str:
    filesystem.write_sys(
        os.path.join(gen_cpu_path(core), 'cpufreq/scaling_governor'),
        data=governor)
    return get_governor(core)
